"""
Translations for email notifications. Strings are keyed by locale (e.g. en_US, fr_FR).
Emails are sent in the recipient's preferred locale when available.

Supported locales: en_US, fr_FR, es_ES, ja_JP, de_DE, nl_NL, zh_CN, pt_BR.
To add a language: copy the en_US block, change the key (e.g. de_DE) and translate
all values. Then add the 2-letter code to _normalize_locale's lang_map if desired.
"""

from zou.app import config

# Default locale used when user locale is missing or unsupported
DEFAULT_EMAIL_LOCALE = "en_US"

# Email translation strings per locale.
# Use %(name)s placeholders for interpolation.
EMAIL_TRANSLATIONS = {
    "en_US": {
        # Comment notification
        "comment_subject": "[Kitsu] %(task_status_name)s - %(author_first_name)s commented on %(task_name)s",
        "comment_title": "New Comment",
        "comment_body_with_text": """<p><strong>%(author_full_name)s</strong> wrote a comment on <a href="%(task_url)s">%(task_name)s</a> and set the status to <strong>%(task_status_name)s</strong>.</p>

<p><em>%(comment_text)s</em></p>
""",
        "comment_body_status_only": """<p><strong>%(author_full_name)s</strong> changed status of <a href="%(task_url)s">%(task_name)s</a> to <strong>%(task_status_name)s</strong>.</p>
""",
        # Mention notification
        "mention_subject": "[Kitsu] %(author_first_name)s mentioned you on %(task_name)s",
        "mention_title": "New Mention",
        "mention_body": """<p><strong>%(author_full_name)s</strong> mentioned you in a comment on <a href="%(task_url)s">%(task_name)s</a>:</p>

<p><em>%(comment_text)s</em></p>
""",
        # Assignation notification
        "assignation_subject": "[Kitsu] You were assigned to %(task_name)s",
        "assignation_title": "New Assignation",
        "assignation_body": """<p><strong>%(author_full_name)s</strong> assigned you to <a href="%(task_url)s">%(task_name)s</a>.</p>
""",
        # Reply notification
        "reply_subject": "[Kitsu] %(author_first_name)s replied on %(task_name)s",
        "reply_title": "New Reply",
        "reply_body": """<p><strong>%(author_full_name)s</strong> wrote a reply on <a href="%(task_url)s">%(task_name)s</a>.</p>

<p><em>%(reply_text)s</em></p>
""",
        # Playlist ready notification
        "playlist_subject": "[Kitsu] The playlist %(playlist_name)s in project %(project_name)s is ready for review",
        "playlist_title": "New Playlist Ready",
        "playlist_body": """<p><strong>%(author_full_name)s</strong> notifies you that playlist <a href="%(playlist_url)s">%(playlist_name)s</a> is ready for a review under %(episode_segment)sthe project %(project_name)s.</p>
""",
        "playlist_episode_segment": "the episode %(episode_name)s of ",
        "playlist_elements_count": "\n<p>%(count)s elements are listed in the playlist.</p>\n",
        "email_signature": "\n<p>Best regards,</p>\n\n<p>%(organisation_name)s Team</p>",
        # Auth: Email OTP
        "auth_otp_subject": "%(organisation_name)s - Kitsu: your verification code",
        "auth_otp_title": "Your verification code",
        "auth_otp_body": """<p>Hello %(first_name)s,</p>

<p>Your verification code is: <strong>%(otp)s</strong></p>

<p>This one time password will expire after 5 minutes. After, you will have to request a new one.
This email was sent at this date: %(time_string)s.
The IP of the person who requested this is: %(person_IP)s.</p>
""",
        # Auth: Password changed
        "auth_password_changed_subject": "%(organisation_name)s - Kitsu: password changed",
        "auth_password_changed_title": "Password Changed",
        "auth_password_changed_body": """<p>Hello %(first_name)s,</p>

<p>You have successfully changed your password at this date: %(time_string)s.</p>
<p>Your IP when you have changed your password is: %(person_IP)s.</p>
""",
        # Auth: Password recovery
        "auth_password_recovery_subject": "%(organisation_name)s - Kitsu: password recovery",
        "auth_password_recovery_title": "Password Recovery",
        "auth_password_recovery_body": """<p>Hello %(first_name)s,</p>

<p>You have requested a password reset. Click on the following button to change your password:</p>

<p class="cta"><a class="button" href="%(reset_url)s">Change your password</a></p>

<p>This link will expire after 2 hours. After, you have to do a new request to reset your password.
This email was sent at this date: %(time_string)s.
The IP of the person who requested this is: %(person_IP)s.</p>
""",
        # Auth: Invitation
        "auth_invitation_subject": "You are invited by %(organisation_name)s to join their Kitsu platform",
        "auth_invitation_title": "Welcome to Kitsu",
        "auth_invitation_body": """<p>Hello %(first_name)s,</p>
<p>You are invited by %(organisation_name)s to join their team on Kitsu.</p>
<p>Your login is: <strong>%(email)s</strong></p>
<p>Set your password to continue:</p>
<p class="cta"><a class="button" href="%(reset_url)s">Set your password</a></p>
""",
        # Auth: Password changed by admin
        "auth_password_changed_by_admin_subject": "%(organisation_name)s - Kitsu: password changed",
        "auth_password_changed_by_admin_title": "Password Changed",
        "auth_password_changed_by_admin_body": """<p>Hello %(first_name)s,</p>
<p>Your password was changed at this date: %(time_string)s.</p>
<p>The IP of the user who changed your password is: %(person_IP)s.</p>
<p>If you don't know the person who changed the password, please contact our support team.</p>
""",
        # Auth: Two factor authentication disabled by admin
        "auth_2fa_disabled_by_admin_subject": "%(organisation_name)s - Kitsu: two factor authentication disabled",
        "auth_2fa_disabled_by_admin_title": "Two Factor Authentication Disabled",
        "auth_2fa_disabled_by_admin_body": """<p>Hello %(first_name)s,</p>
<p>Your two factor authentication was disabled at this date: %(time_string)s.</p>
<p>The IP of the user who disabled your two factor authentication is: %(person_IP)s.</p>
<p>If you don't know the person who disabled the two factor authentication, please contact our support team.</p>
""",
    },
    "fr_FR": {
        # Comment notification
        "comment_subject": "[Kitsu] %(task_status_name)s - %(author_first_name)s a commenté %(task_name)s",
        "comment_title": "Nouveau commentaire",
        "comment_body_with_text": """<p><strong>%(author_full_name)s</strong> a écrit un commentaire sur <a href="%(task_url)s">%(task_name)s</a> et a défini le statut à <strong>%(task_status_name)s</strong>.</p>

<p><em>%(comment_text)s</em></p>
""",
        "comment_body_status_only": """<p><strong>%(author_full_name)s</strong> a changé le statut de <a href="%(task_url)s">%(task_name)s</a> en <strong>%(task_status_name)s</strong>.</p>
""",
        # Mention notification
        "mention_subject": "[Kitsu] %(author_first_name)s vous a mentionné sur %(task_name)s",
        "mention_title": "Nouvelle mention",
        "mention_body": """<p><strong>%(author_full_name)s</strong> vous a mentionné dans un commentaire sur <a href="%(task_url)s">%(task_name)s</a> :</p>

<p><em>%(comment_text)s</em></p>
""",
        # Assignation notification
        "assignation_subject": "[Kitsu] Vous avez été assigné à %(task_name)s",
        "assignation_title": "Nouvelle assignation",
        "assignation_body": """<p><strong>%(author_full_name)s</strong> vous a assigné à <a href="%(task_url)s">%(task_name)s</a>.</p>
""",
        # Reply notification
        "reply_subject": "[Kitsu] %(author_first_name)s a répondu sur %(task_name)s",
        "reply_title": "Nouvelle réponse",
        "reply_body": """<p><strong>%(author_full_name)s</strong> a répondu sur <a href="%(task_url)s">%(task_name)s</a>.</p>

<p><em>%(reply_text)s</em></p>
""",
        # Playlist ready notification
        "playlist_subject": "[Kitsu] La playlist %(playlist_name)s du projet %(project_name)s est prête pour relecture",
        "playlist_title": "Nouvelle playlist prête",
        "playlist_body": """<p><strong>%(author_full_name)s</strong> vous informe que la playlist <a href="%(playlist_url)s">%(playlist_name)s</a> est prête pour relecture dans %(episode_segment)sle projet %(project_name)s.</p>
""",
        "playlist_episode_segment": "l'épisode %(episode_name)s du ",
        "playlist_elements_count": "\n<p>%(count)s éléments sont dans la playlist.</p>\n",
        "email_signature": "\n<p>Cordialement,</p>\n\n<p>L'équipe %(organisation_name)s</p>",
        "auth_otp_subject": "%(organisation_name)s - Kitsu : votre code de vérification",
        "auth_otp_title": "Votre code de vérification",
        "auth_otp_body": """<p>Bonjour %(first_name)s,</p>

<p>Votre code de vérification est : <strong>%(otp)s</strong></p>

<p>Ce mot de passe à usage unique expire dans 5 minutes. Ensuite, vous devrez en demander un nouveau.
Cet e-mail a été envoyé à la date : %(time_string)s.
L'IP de la personne ayant fait la demande est : %(person_IP)s.</p>
""",
        "auth_password_changed_subject": "%(organisation_name)s - Kitsu : mot de passe modifié",
        "auth_password_changed_title": "Mot de passe modifié",
        "auth_password_changed_body": """<p>Bonjour %(first_name)s,</p>

<p>Vous avez modifié votre mot de passe avec succès à la date : %(time_string)s.</p>
<p>Votre adresse IP lors du changement est : %(person_IP)s.</p>
""",
        "auth_password_recovery_subject": "%(organisation_name)s - Kitsu : récupération de mot de passe",
        "auth_password_recovery_title": "Récupération de mot de passe",
        "auth_password_recovery_body": """<p>Bonjour %(first_name)s,</p>

<p>Vous avez demandé une réinitialisation de mot de passe. Cliquez sur le bouton suivant pour le modifier :</p>

<p class="cta"><a class="button" href="%(reset_url)s">Changer votre mot de passe</a></p>

<p>Ce lien expire dans 2 heures. Ensuite, vous devrez faire une nouvelle demande.
Cet e-mail a été envoyé à la date : %(time_string)s.
L'IP de la personne ayant fait la demande est : %(person_IP)s.</p>
""",
        "auth_invitation_subject": "Vous êtes invité par %(organisation_name)s à rejoindre leur plateforme Kitsu",
        "auth_invitation_title": "Bienvenue sur Kitsu",
        "auth_invitation_body": """<p>Bonjour %(first_name)s,</p>
<p>Vous êtes invité par %(organisation_name)s à rejoindre leur équipe sur Kitsu.</p>
<p>Votre identifiant : <strong>%(email)s</strong></p>
<p>Définissez votre mot de passe pour continuer :</p>
<p class="cta"><a class="button" href="%(reset_url)s">Définir votre mot de passe</a></p>
""",
        # Auth: Password changed by admin
        "auth_password_changed_by_admin_subject": "%(organisation_name)s - Kitsu : mot de passe modifié par un administrateur",
        "auth_password_changed_by_admin_title": "Mot de passe modifié",
        "auth_password_changed_by_admin_body": """<p>Bonjour %(first_name)s,</p>
<p>Votre mot de passe a été modifié à cette date : %(time_string)s.</p>
<p>L'IP de la personne qui a modifié votre mot de passe est : %(person_IP)s.</p>
<p>Si vous ne connaissez pas cette personne, veuillez contacter notre équipe de support.</p>
""",
        # Auth: Two factor authentication disabled by admin
        "auth_2fa_disabled_by_admin_subject": "%(organisation_name)s - Kitsu : authentification à deux facteurs désactivée",
        "auth_2fa_disabled_by_admin_title": "Authentification à deux facteurs désactivée",
        "auth_2fa_disabled_by_admin_body": """<p>Bonjour %(first_name)s,</p>
<p>Votre authentification à deux facteurs a été désactivée à cette date : %(time_string)s.</p>
<p>L'IP de la personne qui a désactivé votre authentification à deux facteurs est : %(person_IP)s.</p>
<p>Si vous ne connaissez pas cette personne, veuillez contacter notre équipe de support.</p>
""",
    },
    "es_ES": {
        # Comment notification
        "comment_subject": "[Kitsu] %(task_status_name)s - %(author_first_name)s comentó en %(task_name)s",
        "comment_title": "Nuevo comentario",
        "comment_body_with_text": """<p><strong>%(author_full_name)s</strong> escribió un comentario en <a href="%(task_url)s">%(task_name)s</a> y estableció el estado en <strong>%(task_status_name)s</strong>.</p>

<p><em>%(comment_text)s</em></p>
""",
        "comment_body_status_only": """<p><strong>%(author_full_name)s</strong> cambió el estado de <a href="%(task_url)s">%(task_name)s</a> a <strong>%(task_status_name)s</strong>.</p>
""",
        # Mention notification
        "mention_subject": "[Kitsu] %(author_first_name)s te mencionó en %(task_name)s",
        "mention_title": "Nueva mención",
        "mention_body": """<p><strong>%(author_full_name)s</strong> te mencionó en un comentario en <a href="%(task_url)s">%(task_name)s</a>:</p>

<p><em>%(comment_text)s</em></p>
""",
        # Assignation notification
        "assignation_subject": "[Kitsu] Has sido asignado a %(task_name)s",
        "assignation_title": "Nueva asignación",
        "assignation_body": """<p><strong>%(author_full_name)s</strong> te asignó a <a href="%(task_url)s">%(task_name)s</a>.</p>
""",
        # Reply notification
        "reply_subject": "[Kitsu] %(author_first_name)s respondió en %(task_name)s",
        "reply_title": "Nueva respuesta",
        "reply_body": """<p><strong>%(author_full_name)s</strong> respondió en <a href="%(task_url)s">%(task_name)s</a>.</p>

<p><em>%(reply_text)s</em></p>
""",
        # Playlist ready notification
        "playlist_subject": "[Kitsu] La playlist %(playlist_name)s del proyecto %(project_name)s está lista para revisión",
        "playlist_title": "Nueva playlist lista",
        "playlist_body": """<p><strong>%(author_full_name)s</strong> te informa que la playlist <a href="%(playlist_url)s">%(playlist_name)s</a> está lista para revisión en %(episode_segment)sel proyecto %(project_name)s.</p>
""",
        "playlist_episode_segment": "el episodio %(episode_name)s de ",
        "playlist_elements_count": "\n<p>%(count)s elementos en la playlist.</p>\n",
        "email_signature": "\n<p>Saludos cordiales,</p>\n\n<p>El equipo de %(organisation_name)s</p>",
        "auth_otp_subject": "%(organisation_name)s - Kitsu: tu código de verificación",
        "auth_otp_title": "Tu código de verificación",
        "auth_otp_body": """<p>Hola %(first_name)s,</p>

<p>Tu código de verificación es: <strong>%(otp)s</strong></p>

<p>Esta contraseña de un solo uso caduca en 5 minutos. Después tendrás que solicitar una nueva.
Este correo se envió en esta fecha: %(time_string)s.
La IP de la persona que lo solicitó es: %(person_IP)s.</p>
""",
        "auth_password_changed_subject": "%(organisation_name)s - Kitsu: contraseña cambiada",
        "auth_password_changed_title": "Contraseña cambiada",
        "auth_password_changed_body": """<p>Hola %(first_name)s,</p>

<p>Has cambiado tu contraseña correctamente en esta fecha: %(time_string)s.</p>
<p>Tu IP al cambiar la contraseña es: %(person_IP)s.</p>
""",
        "auth_password_recovery_subject": "%(organisation_name)s - Kitsu: recuperación de contraseña",
        "auth_password_recovery_title": "Recuperación de contraseña",
        "auth_password_recovery_body": """<p>Hola %(first_name)s,</p>

<p>Has solicitado restablecer tu contraseña. Haz clic en el siguiente botón para cambiarla:</p>

<p class="cta"><a class="button" href="%(reset_url)s">Cambiar tu contraseña</a></p>

<p>Este enlace caduca en 2 horas. Después tendrás que solicitar un nuevo restablecimiento.
Este correo se envió en esta fecha: %(time_string)s.
La IP de la persona que lo solicitó es: %(person_IP)s.</p>
""",
        "auth_invitation_subject": "Has sido invitado por %(organisation_name)s a unirte a su plataforma Kitsu",
        "auth_invitation_title": "Bienvenido a Kitsu",
        "auth_invitation_body": """<p>Hola %(first_name)s,</p>
<p>Has sido invitado por %(organisation_name)s a unirte a su equipo en Kitsu.</p>
<p>Tu inicio de sesión es: <strong>%(email)s</strong></p>
<p>Establece tu contraseña para continuar:</p>
<p class="cta"><a class="button" href="%(reset_url)s">Establecer tu contraseña</a></p>
""",
        # Auth: Password changed by admin
        "auth_password_changed_by_admin_subject": "%(organisation_name)s - Kitsu: contraseña cambiada por un administrador",
        "auth_password_changed_by_admin_title": "Contraseña cambiada",
        "auth_password_changed_by_admin_body": """<p>Hola %(first_name)s,</p>
<p>Tu contraseña fue cambiada en esta fecha: %(time_string)s.</p>
<p>La IP de la persona que cambió tu contraseña es: %(person_IP)s.</p>
<p>Si no conoces a esta persona, por favor contacta a nuestro equipo de soporte.</p>
""",
        # Auth: Two factor authentication disabled by admin
        "auth_2fa_disabled_by_admin_subject": "%(organisation_name)s - Kitsu: autenticación de dos factores desactivada",
        "auth_2fa_disabled_by_admin_title": "Autenticación de dos factores desactivada",
        "auth_2fa_disabled_by_admin_body": """<p>Hola %(first_name)s,</p>
<p>Tu autenticación de dos factores fue desactivada en esta fecha: %(time_string)s.</p>
<p>La IP de la persona que desactivó tu autenticación de dos factores es: %(person_IP)s.</p>
<p>Si no conoces a esta persona, por favor contacta a nuestro equipo de soporte.</p>
""",
    },
    "ja_JP": {
        # Comment notification
        "comment_subject": "[Kitsu] %(task_status_name)s - %(author_first_name)s が %(task_name)s にコメントしました",
        "comment_title": "新しいコメント",
        "comment_body_with_text": """<p><strong>%(author_full_name)s</strong> が <a href="%(task_url)s">%(task_name)s</a> にコメントし、ステータスを <strong>%(task_status_name)s</strong> に設定しました。</p>

<p><em>%(comment_text)s</em></p>
""",
        "comment_body_status_only": """<p><strong>%(author_full_name)s</strong> が <a href="%(task_url)s">%(task_name)s</a> のステータスを <strong>%(task_status_name)s</strong> に変更しました。</p>
""",
        # Mention notification
        "mention_subject": "[Kitsu] %(author_first_name)s が %(task_name)s であなたをメンションしました",
        "mention_title": "新しいメンション",
        "mention_body": """<p><strong>%(author_full_name)s</strong> が <a href="%(task_url)s">%(task_name)s</a> のコメントであなたをメンションしました：</p>

<p><em>%(comment_text)s</em></p>
""",
        # Assignation notification
        "assignation_subject": "[Kitsu] %(task_name)s にアサインされました",
        "assignation_title": "新しいアサイン",
        "assignation_body": """<p><strong>%(author_full_name)s</strong> があなたを <a href="%(task_url)s">%(task_name)s</a> にアサインしました。</p>
""",
        # Reply notification
        "reply_subject": "[Kitsu] %(author_first_name)s が %(task_name)s に返信しました",
        "reply_title": "新しい返信",
        "reply_body": """<p><strong>%(author_full_name)s</strong> が <a href="%(task_url)s">%(task_name)s</a> に返信しました。</p>

<p><em>%(reply_text)s</em></p>
""",
        # Playlist ready notification
        "playlist_subject": "[Kitsu] プロジェクト %(project_name)s のプレイリスト %(playlist_name)s がレビュー可能です",
        "playlist_title": "新しいプレイリストの準備ができました",
        "playlist_body": """<p><strong>%(author_full_name)s</strong> より、プレイリスト <a href="%(playlist_url)s">%(playlist_name)s</a> が %(episode_segment)sプロジェクト %(project_name)s でレビュー可能になったとのお知らせです。</p>
""",
        "playlist_episode_segment": "エピソード %(episode_name)s の ",
        "playlist_elements_count": "\n<p>プレイリストに %(count)s 件の要素が含まれています。</p>\n",
        "email_signature": "\n<p>よろしくお願いいたします。</p>\n\n<p>%(organisation_name)s チーム</p>",
        "auth_otp_subject": "%(organisation_name)s - Kitsu: 認証コード",
        "auth_otp_title": "認証コード",
        "auth_otp_body": """<p>%(first_name)s 様</p>

<p>認証コードは <strong>%(otp)s</strong> です。</p>

<p>このワンタイムパスワードは5分で失効します。その後は新しいコードをリクエストしてください。
送信日時: %(time_string)s。
リクエスト元のIP: %(person_IP)s。</p>
""",
        "auth_password_changed_subject": "%(organisation_name)s - Kitsu: パスワードを変更しました",
        "auth_password_changed_title": "パスワード変更完了",
        "auth_password_changed_body": """<p>%(first_name)s 様</p>

<p>パスワードを正常に変更しました。変更日時: %(time_string)s。</p>
<p>変更時のIP: %(person_IP)s。</p>
""",
        "auth_password_recovery_subject": "%(organisation_name)s - Kitsu: パスワードのリセット",
        "auth_password_recovery_title": "パスワードのリセット",
        "auth_password_recovery_body": """<p>%(first_name)s 様</p>

<p>パスワードのリセットをリクエストされました。以下のボタンからパスワードを変更してください。</p>

<p class="cta"><a class="button" href="%(reset_url)s">パスワードを変更</a></p>

<p>このリンクは2時間で失効します。その後は再度リセットをリクエストしてください。
送信日時: %(time_string)s。
リクエスト元のIP: %(person_IP)s。</p>
""",
        "auth_invitation_subject": "%(organisation_name)s から Kitsu への招待",
        "auth_invitation_title": "Kitsu へようこそ",
        "auth_invitation_body": """<p>%(first_name)s 様</p>
<p>%(organisation_name)s より Kitsu のチームに参加する招待が届きました。</p>
<p>ログイン: <strong>%(email)s</strong></p>
<p>続行するにはパスワードを設定してください:</p>
<p class="cta"><a class="button" href="%(reset_url)s">パスワードを設定</a></p>
""",
        # Auth: Password changed by admin
        "auth_password_changed_by_admin_subject": "%(organisation_name)s - Kitsu: 管理者によりパスワードが変更されました",
        "auth_password_changed_by_admin_title": "パスワード変更完了",
        "auth_password_changed_by_admin_body": """<p>%(first_name)s 様</p>
<p>あなたのパスワードが次の日時に変更されました: %(time_string)s。</p>
<p>パスワードを変更した人のIPアドレス: %(person_IP)s。</p>
<p>心当たりがない場合は、サポートチームにご連絡ください。</p>
""",
        # Auth: Two factor authentication disabled by admin
        "auth_2fa_disabled_by_admin_subject": "%(organisation_name)s - Kitsu: 管理者により二要素認証が無効化されました",
        "auth_2fa_disabled_by_admin_title": "二要素認証が無効化されました",
        "auth_2fa_disabled_by_admin_body": """<p>%(first_name)s 様</p>
<p>あなたの二要素認証が次の日時に無効化されました: %(time_string)s。</p>
<p>二要素認証を無効化した人のIPアドレス: %(person_IP)s。</p>
<p>心当たりがない場合は、サポートチームにご連絡ください。</p>
""",
    },
    "de_DE": {
        "comment_subject": "[Kitsu] %(task_status_name)s - %(author_first_name)s hat %(task_name)s kommentiert",
        "comment_title": "Neuer Kommentar",
        "comment_body_with_text": """<p><strong>%(author_full_name)s</strong> hat einen Kommentar zu <a href="%(task_url)s">%(task_name)s</a> geschrieben und den Status auf <strong>%(task_status_name)s</strong> gesetzt.</p>

<p><em>%(comment_text)s</em></p>
""",
        "comment_body_status_only": """<p><strong>%(author_full_name)s</strong> hat den Status von <a href="%(task_url)s">%(task_name)s</a> auf <strong>%(task_status_name)s</strong> geändert.</p>
""",
        "mention_subject": "[Kitsu] %(author_first_name)s hat Sie in %(task_name)s erwähnt",
        "mention_title": "Neue Erwähnung",
        "mention_body": """<p><strong>%(author_full_name)s</strong> hat Sie in einem Kommentar zu <a href="%(task_url)s">%(task_name)s</a> erwähnt:</p>

<p><em>%(comment_text)s</em></p>
""",
        "assignation_subject": "[Kitsu] Sie wurden %(task_name)s zugewiesen",
        "assignation_title": "Neue Zuweisung",
        "assignation_body": """<p><strong>%(author_full_name)s</strong> hat Sie <a href="%(task_url)s">%(task_name)s</a> zugewiesen.</p>
""",
        "reply_subject": "[Kitsu] %(author_first_name)s hat auf %(task_name)s geantwortet",
        "reply_title": "Neue Antwort",
        "reply_body": """<p><strong>%(author_full_name)s</strong> hat auf <a href="%(task_url)s">%(task_name)s</a> geantwortet.</p>

<p><em>%(reply_text)s</em></p>
""",
        "playlist_subject": "[Kitsu] Die Playlist %(playlist_name)s im Projekt %(project_name)s ist zur Überprüfung bereit",
        "playlist_title": "Neue Playlist bereit",
        "playlist_body": """<p><strong>%(author_full_name)s</strong> teilt mit, dass die Playlist <a href="%(playlist_url)s">%(playlist_name)s</a> zur Überprüfung in %(episode_segment)sdem Projekt %(project_name)s bereit ist.</p>
""",
        "playlist_episode_segment": "der Episode %(episode_name)s von ",
        "playlist_elements_count": "\n<p>%(count)s Elemente sind in der Playlist enthalten.</p>\n",
        "email_signature": "\n<p>Mit freundlichen Grüßen,</p>\n\n<p>%(organisation_name)s Team</p>",
        "auth_otp_subject": "%(organisation_name)s - Kitsu: Ihr Bestätigungscode",
        "auth_otp_title": "Ihr Bestätigungscode",
        "auth_otp_body": """<p>Hallo %(first_name)s,</p>

<p>Ihr Bestätigungscode lautet: <strong>%(otp)s</strong></p>

<p>Dieses Einmalpasswort läuft nach 5 Minuten ab. Danach müssen Sie einen neuen anfordern.
Diese E-Mail wurde am %(time_string)s gesendet.
Die IP der anfragenden Person ist: %(person_IP)s.</p>
""",
        "auth_password_changed_subject": "%(organisation_name)s - Kitsu: Passwort geändert",
        "auth_password_changed_title": "Passwort geändert",
        "auth_password_changed_body": """<p>Hallo %(first_name)s,</p>

<p>Sie haben Ihr Passwort erfolgreich am %(time_string)s geändert.</p>
<p>Ihre IP bei der Änderung war: %(person_IP)s.</p>
""",
        "auth_password_recovery_subject": "%(organisation_name)s - Kitsu: Passwort-Wiederherstellung",
        "auth_password_recovery_title": "Passwort-Wiederherstellung",
        "auth_password_recovery_body": """<p>Hallo %(first_name)s,</p>

<p>Sie haben eine Passwort-Zurücksetzung angefordert. Klicken Sie auf die Schaltfläche unten, um Ihr Passwort zu ändern:</p>

<p class="cta"><a class="button" href="%(reset_url)s">Passwort ändern</a></p>

<p>Dieser Link läuft nach 2 Stunden ab. Danach müssen Sie eine neue Anfrage stellen.
Diese E-Mail wurde am %(time_string)s gesendet.
Die IP der anfragenden Person ist: %(person_IP)s.</p>
""",
        "auth_invitation_subject": "Sie wurden von %(organisation_name)s eingeladen, der Kitsu-Plattform beizutreten",
        "auth_invitation_title": "Willkommen bei Kitsu",
        "auth_invitation_body": """<p>Hallo %(first_name)s,</p>
<p>Sie wurden von %(organisation_name)s eingeladen, ihrem Team auf Kitsu beizutreten.</p>
<p>Ihre Anmeldung: <strong>%(email)s</strong></p>
<p>Setzen Sie Ihr Passwort, um fortzufahren:</p>
<p class="cta"><a class="button" href="%(reset_url)s">Passwort setzen</a></p>
""",
        # Auth: Password changed by admin
        "auth_password_changed_by_admin_subject": "%(organisation_name)s - Kitsu: Passwort von einem Administrator geändert",
        "auth_password_changed_by_admin_title": "Passwort geändert",
        "auth_password_changed_by_admin_body": """<p>Hallo %(first_name)s,</p>
<p>Ihr Passwort wurde zu diesem Zeitpunkt geändert: %(time_string)s.</p>
<p>Die IP der Person, die Ihr Passwort geändert hat, ist: %(person_IP)s.</p>
<p>Wenn Sie diese Person nicht kennen, kontaktieren Sie bitte unser Support-Team.</p>
""",
        # Auth: Two factor authentication disabled by admin
        "auth_2fa_disabled_by_admin_subject": "%(organisation_name)s - Kitsu: Zwei-Faktor-Authentifizierung deaktiviert",
        "auth_2fa_disabled_by_admin_title": "Zwei-Faktor-Authentifizierung deaktiviert",
        "auth_2fa_disabled_by_admin_body": """<p>Hallo %(first_name)s,</p>
<p>Ihre Zwei-Faktor-Authentifizierung wurde zu diesem Zeitpunkt deaktiviert: %(time_string)s.</p>
<p>Die IP der Person, die Ihre Zwei-Faktor-Authentifizierung deaktiviert hat, ist: %(person_IP)s.</p>
<p>Wenn Sie diese Person nicht kennen, kontaktieren Sie bitte unser Support-Team.</p>
""",
    },
    "nl_NL": {
        "comment_subject": "[Kitsu] %(task_status_name)s - %(author_first_name)s heeft gereageerd op %(task_name)s",
        "comment_title": "Nieuwe reactie",
        "comment_body_with_text": """<p><strong>%(author_full_name)s</strong> heeft een reactie geplaatst op <a href="%(task_url)s">%(task_name)s</a> en de status gezet op <strong>%(task_status_name)s</strong>.</p>

<p><em>%(comment_text)s</em></p>
""",
        "comment_body_status_only": """<p><strong>%(author_full_name)s</strong> heeft de status van <a href="%(task_url)s">%(task_name)s</a> gewijzigd naar <strong>%(task_status_name)s</strong>.</p>
""",
        "mention_subject": "[Kitsu] %(author_first_name)s heeft je vermeld in %(task_name)s",
        "mention_title": "Nieuwe vermelding",
        "mention_body": """<p><strong>%(author_full_name)s</strong> heeft je vermeld in een reactie op <a href="%(task_url)s">%(task_name)s</a>:</p>

<p><em>%(comment_text)s</em></p>
""",
        "assignation_subject": "[Kitsu] Je bent toegewezen aan %(task_name)s",
        "assignation_title": "Nieuwe toewijzing",
        "assignation_body": """<p><strong>%(author_full_name)s</strong> heeft je toegewezen aan <a href="%(task_url)s">%(task_name)s</a>.</p>
""",
        "reply_subject": "[Kitsu] %(author_first_name)s heeft geantwoord op %(task_name)s",
        "reply_title": "Nieuw antwoord",
        "reply_body": """<p><strong>%(author_full_name)s</strong> heeft geantwoord op <a href="%(task_url)s">%(task_name)s</a>.</p>

<p><em>%(reply_text)s</em></p>
""",
        "playlist_subject": "[Kitsu] De playlist %(playlist_name)s in project %(project_name)s is klaar voor beoordeling",
        "playlist_title": "Nieuwe playlist klaar",
        "playlist_body": """<p><strong>%(author_full_name)s</strong> meldt dat de playlist <a href="%(playlist_url)s">%(playlist_name)s</a> klaar is voor beoordeling binnen %(episode_segment)shet project %(project_name)s.</p>
""",
        "playlist_episode_segment": "de aflevering %(episode_name)s van ",
        "playlist_elements_count": "\n<p>%(count)s elementen staan in de playlist.</p>\n",
        "email_signature": "\n<p>Met vriendelijke groet,</p>\n\n<p>%(organisation_name)s Team</p>",
        "auth_otp_subject": "%(organisation_name)s - Kitsu: uw verificatiecode",
        "auth_otp_title": "Uw verificatiecode",
        "auth_otp_body": """<p>Hallo %(first_name)s,</p>

<p>Uw verificatiecode is: <strong>%(otp)s</strong></p>

<p>Dit eenmalige wachtwoord verloopt na 5 minuten. Daarna moet u een nieuwe aanvragen.
Deze e-mail is verzonden op: %(time_string)s.
Het IP-adres van de aanvrager is: %(person_IP)s.</p>
""",
        "auth_password_changed_subject": "%(organisation_name)s - Kitsu: wachtwoord gewijzigd",
        "auth_password_changed_title": "Wachtwoord gewijzigd",
        "auth_password_changed_body": """<p>Hallo %(first_name)s,</p>

<p>U heeft uw wachtwoord succesvol gewijzigd op: %(time_string)s.</p>
<p>Uw IP-adres bij de wijziging was: %(person_IP)s.</p>
""",
        "auth_password_recovery_subject": "%(organisation_name)s - Kitsu: wachtwoordherstel",
        "auth_password_recovery_title": "Wachtwoordherstel",
        "auth_password_recovery_body": """<p>Hallo %(first_name)s,</p>

<p>U heeft een wachtwoordreset aangevraagd. Klik op de onderstaande knop om uw wachtwoord te wijzigen:</p>

<p class="cta"><a class="button" href="%(reset_url)s">Wachtwoord wijzigen</a></p>

<p>Deze link verloopt na 2 uur. Daarna moet u een nieuwe aanvraag doen.
Deze e-mail is verzonden op: %(time_string)s.
Het IP-adres van de aanvrager is: %(person_IP)s.</p>
""",
        "auth_invitation_subject": "U bent door %(organisation_name)s uitgenodigd om deel te nemen aan hun Kitsu-platform",
        "auth_invitation_title": "Welkom bij Kitsu",
        "auth_invitation_body": """<p>Hallo %(first_name)s,</p>
<p>U bent door %(organisation_name)s uitgenodigd om deel te nemen aan hun team op Kitsu.</p>
<p>Uw inloggegevens: <strong>%(email)s</strong></p>
<p>Stel uw wachtwoord in om door te gaan:</p>
<p class="cta"><a class="button" href="%(reset_url)s">Wachtwoord instellen</a></p>
""",
        # Auth: Password changed by admin
        "auth_password_changed_by_admin_subject": "%(organisation_name)s - Kitsu: wachtwoord gewijzigd door een beheerder",
        "auth_password_changed_by_admin_title": "Wachtwoord gewijzigd",
        "auth_password_changed_by_admin_body": """<p>Hallo %(first_name)s,</p>
<p>Uw wachtwoord is gewijzigd op: %(time_string)s.</p>
<p>Het IP-adres van de persoon die uw wachtwoord heeft gewijzigd is: %(person_IP)s.</p>
<p>Als u deze persoon niet kent, neem dan contact op met ons supportteam.</p>
""",
        # Auth: Two factor authentication disabled by admin
        "auth_2fa_disabled_by_admin_subject": "%(organisation_name)s - Kitsu: tweefactorauthenticatie uitgeschakeld",
        "auth_2fa_disabled_by_admin_title": "Tweefactorauthenticatie uitgeschakeld",
        "auth_2fa_disabled_by_admin_body": """<p>Hallo %(first_name)s,</p>
<p>Uw tweefactorauthenticatie is uitgeschakeld op: %(time_string)s.</p>
<p>Het IP-adres van de persoon die uw tweefactorauthenticatie heeft uitgeschakeld is: %(person_IP)s.</p>
<p>Als u deze persoon niet kent, neem dan contact op met ons supportteam.</p>
""",
    },
    "zh_CN": {
        "comment_subject": "[Kitsu] %(task_status_name)s - %(author_first_name)s 在 %(task_name)s 上发表了评论",
        "comment_title": "新评论",
        "comment_body_with_text": """<p><strong>%(author_full_name)s</strong> 在 <a href="%(task_url)s">%(task_name)s</a> 上发表了评论，并将状态设为 <strong>%(task_status_name)s</strong>。</p>

<p><em>%(comment_text)s</em></p>
""",
        "comment_body_status_only": """<p><strong>%(author_full_name)s</strong> 将 <a href="%(task_url)s">%(task_name)s</a> 的状态改为 <strong>%(task_status_name)s</strong>。</p>
""",
        "mention_subject": "[Kitsu] %(author_first_name)s 在 %(task_name)s 中提到了您",
        "mention_title": "新@提及",
        "mention_body": """<p><strong>%(author_full_name)s</strong> 在 <a href="%(task_url)s">%(task_name)s</a> 的评论中提到了您：</p>

<p><em>%(comment_text)s</em></p>
""",
        "assignation_subject": "[Kitsu] 您已被分配至 %(task_name)s",
        "assignation_title": "新分配",
        "assignation_body": """<p><strong>%(author_full_name)s</strong> 已将您分配至 <a href="%(task_url)s">%(task_name)s</a>。</p>
""",
        "reply_subject": "[Kitsu] %(author_first_name)s 在 %(task_name)s 上回复了",
        "reply_title": "新回复",
        "reply_body": """<p><strong>%(author_full_name)s</strong> 在 <a href="%(task_url)s">%(task_name)s</a> 上回复了。</p>

<p><em>%(reply_text)s</em></p>
""",
        "playlist_subject": "[Kitsu] 项目 %(project_name)s 中的播放列表 %(playlist_name)s 已可审核",
        "playlist_title": "新播放列表已就绪",
        "playlist_body": """<p><strong>%(author_full_name)s</strong> 通知您，播放列表 <a href="%(playlist_url)s">%(playlist_name)s</a> 已在 %(episode_segment)s项目 %(project_name)s 中准备好供审核。</p>
""",
        "playlist_episode_segment": "第 %(episode_name)s 集 ",
        "playlist_elements_count": "\n<p>播放列表中共 %(count)s 个元素。</p>\n",
        "email_signature": "\n<p>此致</p>\n\n<p>%(organisation_name)s 团队</p>",
        "auth_otp_subject": "%(organisation_name)s - Kitsu：您的验证码",
        "auth_otp_title": "您的验证码",
        "auth_otp_body": """<p>您好 %(first_name)s，</p>

<p>您的验证码是：<strong>%(otp)s</strong></p>

<p>此一次性密码将在 5 分钟后失效，之后您需要重新请求。
此邮件发送时间：%(time_string)s。
请求者 IP：%(person_IP)s。</p>
""",
        "auth_password_changed_subject": "%(organisation_name)s - Kitsu：密码已修改",
        "auth_password_changed_title": "密码已修改",
        "auth_password_changed_body": """<p>您好 %(first_name)s，</p>

<p>您已成功于 %(time_string)s 修改密码。</p>
<p>修改时的 IP：%(person_IP)s。</p>
""",
        "auth_password_recovery_subject": "%(organisation_name)s - Kitsu：找回密码",
        "auth_password_recovery_title": "找回密码",
        "auth_password_recovery_body": """<p>您好 %(first_name)s，</p>

<p>您已请求重置密码。请点击下方按钮修改密码：</p>

<p class="cta"><a class="button" href="%(reset_url)s">修改密码</a></p>

<p>此链接将在 2 小时后失效，之后需重新请求。
此邮件发送时间：%(time_string)s。
请求者 IP：%(person_IP)s。</p>
""",
        "auth_invitation_subject": "%(organisation_name)s 邀请您加入 Kitsu 平台",
        "auth_invitation_title": "欢迎使用 Kitsu",
        "auth_invitation_body": """<p>您好 %(first_name)s，</p>
<p>%(organisation_name)s 邀请您加入他们在 Kitsu 的团队。</p>
<p>您的登录名：<strong>%(email)s</strong></p>
<p>请设置密码以继续：</p>
<p class="cta"><a class="button" href="%(reset_url)s">设置密码</a></p>
""",
        # Auth: Password changed by admin
        "auth_password_changed_by_admin_subject": "%(organisation_name)s - Kitsu：管理员已更改密码",
        "auth_password_changed_by_admin_title": "密码已更改",
        "auth_password_changed_by_admin_body": """<p>您好 %(first_name)s，</p>
<p>您的密码已于此日期更改：%(time_string)s。</p>
<p>更改您密码的人的 IP 地址为：%(person_IP)s。</p>
<p>如果您不认识此人，请联系我们的支持团队。</p>
""",
        # Auth: Two factor authentication disabled by admin
        "auth_2fa_disabled_by_admin_subject": "%(organisation_name)s - Kitsu：双因素认证已被管理员禁用",
        "auth_2fa_disabled_by_admin_title": "双因素认证已禁用",
        "auth_2fa_disabled_by_admin_body": """<p>您好 %(first_name)s，</p>
<p>您的双因素认证已于此日期被禁用：%(time_string)s。</p>
<p>禁用您双因素认证的人的 IP 地址为：%(person_IP)s。</p>
<p>如果您不认识此人，请联系我们的支持团队。</p>
""",
    },
    "pt_BR": {
        "comment_subject": "[Kitsu] %(task_status_name)s - %(author_first_name)s comentou em %(task_name)s",
        "comment_title": "Novo comentário",
        "comment_body_with_text": """<p><strong>%(author_full_name)s</strong> escreveu um comentário em <a href="%(task_url)s">%(task_name)s</a> e definiu o status como <strong>%(task_status_name)s</strong>.</p>

<p><em>%(comment_text)s</em></p>
""",
        "comment_body_status_only": """<p><strong>%(author_full_name)s</strong> alterou o status de <a href="%(task_url)s">%(task_name)s</a> para <strong>%(task_status_name)s</strong>.</p>
""",
        "mention_subject": "[Kitsu] %(author_first_name)s mencionou você em %(task_name)s",
        "mention_title": "Nova menção",
        "mention_body": """<p><strong>%(author_full_name)s</strong> mencionou você em um comentário em <a href="%(task_url)s">%(task_name)s</a>:</p>

<p><em>%(comment_text)s</em></p>
""",
        "assignation_subject": "[Kitsu] Você foi atribuído a %(task_name)s",
        "assignation_title": "Nova atribuição",
        "assignation_body": """<p><strong>%(author_full_name)s</strong> atribuiu você a <a href="%(task_url)s">%(task_name)s</a>.</p>
""",
        "reply_subject": "[Kitsu] %(author_first_name)s respondeu em %(task_name)s",
        "reply_title": "Nova resposta",
        "reply_body": """<p><strong>%(author_full_name)s</strong> respondeu em <a href="%(task_url)s">%(task_name)s</a>.</p>

<p><em>%(reply_text)s</em></p>
""",
        "playlist_subject": "[Kitsu] A playlist %(playlist_name)s no projeto %(project_name)s está pronta para revisão",
        "playlist_title": "Nova playlist pronta",
        "playlist_body": """<p><strong>%(author_full_name)s</strong> informa que a playlist <a href="%(playlist_url)s">%(playlist_name)s</a> está pronta para revisão em %(episode_segment)so projeto %(project_name)s.</p>
""",
        "playlist_episode_segment": "o episódio %(episode_name)s de ",
        "playlist_elements_count": "\n<p>%(count)s elementos na playlist.</p>\n",
        "email_signature": "\n<p>Atenciosamente,</p>\n\n<p>Equipe %(organisation_name)s</p>",
        "auth_otp_subject": "%(organisation_name)s - Kitsu: seu código de verificação",
        "auth_otp_title": "Seu código de verificação",
        "auth_otp_body": """<p>Olá %(first_name)s,</p>

<p>Seu código de verificação é: <strong>%(otp)s</strong></p>

<p>Esta senha de uso único expira em 5 minutos. Depois, você precisará solicitar uma nova.
Este e-mail foi enviado em: %(time_string)s.
O IP de quem solicitou é: %(person_IP)s.</p>
""",
        "auth_password_changed_subject": "%(organisation_name)s - Kitsu: senha alterada",
        "auth_password_changed_title": "Senha alterada",
        "auth_password_changed_body": """<p>Olá %(first_name)s,</p>

<p>Você alterou sua senha com sucesso em: %(time_string)s.</p>
<p>Seu IP ao alterar a senha foi: %(person_IP)s.</p>
""",
        "auth_password_recovery_subject": "%(organisation_name)s - Kitsu: recuperação de senha",
        "auth_password_recovery_title": "Recuperação de senha",
        "auth_password_recovery_body": """<p>Olá %(first_name)s,</p>

<p>Você solicitou a redefinição de senha. Clique no botão abaixo para alterá-la:</p>

<p class="cta"><a class="button" href="%(reset_url)s">Alterar sua senha</a></p>

<p>Este link expira em 2 horas. Depois, você precisará fazer uma nova solicitação.
Este e-mail foi enviado em: %(time_string)s.
O IP de quem solicitou é: %(person_IP)s.</p>
""",
        "auth_invitation_subject": "Você foi convidado por %(organisation_name)s para participar da plataforma Kitsu",
        "auth_invitation_title": "Bem-vindo ao Kitsu",
        "auth_invitation_body": """<p>Olá %(first_name)s,</p>
<p>Você foi convidado por %(organisation_name)s para participar da equipe no Kitsu.</p>
<p>Seu login: <strong>%(email)s</strong></p>
<p>Defina sua senha para continuar:</p>
<p class="cta"><a class="button" href="%(reset_url)s">Definir sua senha</a></p>
""",
        # Auth: Password changed by admin
        "auth_password_changed_by_admin_subject": "%(organisation_name)s - Kitsu: senha alterada por um administrador",
        "auth_password_changed_by_admin_title": "Senha alterada",
        "auth_password_changed_by_admin_body": """<p>Olá %(first_name)s,</p>
<p>Sua senha foi alterada nesta data: %(time_string)s.</p>
<p>O IP da pessoa que alterou sua senha é: %(person_IP)s.</p>
<p>Se você não conhece esta pessoa, entre em contato com nossa equipe de suporte.</p>
""",
        # Auth: Two factor authentication disabled by admin
        "auth_2fa_disabled_by_admin_subject": "%(organisation_name)s - Kitsu: autenticação de dois fatores desativada",
        "auth_2fa_disabled_by_admin_title": "Autenticação de dois fatores desativada",
        "auth_2fa_disabled_by_admin_body": """<p>Olá %(first_name)s,</p>
<p>Sua autenticação de dois fatores foi desativada nesta data: %(time_string)s.</p>
<p>O IP da pessoa que desativou sua autenticação de dois fatores é: %(person_IP)s.</p>
<p>Se você não conhece esta pessoa, entre em contato com nossa equipe de suporte.</p>
""",
    },
}


def _normalize_locale(locale):
    """
    Return a locale string suitable for lookup (e.g. en_US, fr_FR).
    """
    if not locale or not isinstance(locale, str):
        locale = "en_US"

    locale = locale.strip()
    if len(locale) == 2:
        lang_map = {
            "en": "en_US",
            "fr": "fr_FR",
            "es": "es_ES",
            "ja": "ja_JP",
            "de": "de_DE",
            "nl": "nl_NL",
            "zh": "zh_CN",
            "pt": "pt_BR",
        }
        return lang_map.get(locale.lower(), locale)
    return locale


def get_email_translation(locale, key, **params):
    """
    Return the translated string for the given locale and key.
    Interpolates params with %(name)s placeholders.

    Fallback order if the locale is missing or not supported:
    1. User's locale (e.g. fr_FR, ja_JP)
    2. en_US (default English)
    3. config.DEFAULT_LOCALE (usually en_US)
    """
    normalized = _normalize_locale(locale)
    fallback_chain = (
        normalized,
        "en_US",  # Always fall back to US English if locale unsupported
        getattr(config, "DEFAULT_LOCALE", "en_US"),
    )
    for candidate in fallback_chain:
        if not candidate:
            continue
        trans = EMAIL_TRANSLATIONS.get(candidate)
        if trans and key in trans:
            try:
                return trans[key] % params
            except (KeyError, TypeError):
                return trans[key]
    # Last resort: use key as string (should not happen if en_US is complete)
    return key



"""add frontend options to plugins

Revision ID: 9a9df20ea5a7
Revises: 12208e50bf18
Create Date: 2025-12-23 00:47:50.621868

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils


# revision identifiers, used by Alembic.
revision = "9a9df20ea5a7"
down_revision = "12208e50bf18"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("plugin", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("frontend_project_enabled", sa.Boolean(), nullable=True)
        )
        batch_op.add_column(
            sa.Column("frontend_studio_enabled", sa.Boolean(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("plugin", schema=None) as batch_op:
        batch_op.drop_column("frontend_studio_enabled")
        batch_op.drop_column("frontend_project_enabled")

    # ### end Alembic commands ###



from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.admin.resources import (
    ConfigCheckResource,
)

routes = [
    ("/admin/config/check", ConfigCheckResource),
]

blueprint = Blueprint("admin", "admin")
api = configure_api_from_blueprint(blueprint, routes)



import sys

import redis

from zou.app import config

USER_LIMIT_KEY = "config:user_limit"
DEFAULT_TIMEZONE_KEY = "config:default_timezone"
DEFAULT_LOCALE_KEY = "config:default_locale"
NOMAD_HOST_KEY = "config:nomad_host"
NOMAD_NORMALIZE_JOB_KEY = "config:nomad_normalize_job"
NOMAD_PLAYLIST_JOB_KEY = "config:nomad_playlist_job"

try:
    config_store = redis.StrictRedis(
        host=config.KEY_VALUE_STORE["host"],
        port=config.KEY_VALUE_STORE["port"],
        db=config.KV_CONFIG_DB_INDEX,
        password=config.KEY_VALUE_STORE["password"],
        decode_responses=True,
    )
    config_store.ping()
except redis.ConnectionError:
    config_store = None
    if "pytest" not in sys.modules:
        print("Cannot access to the required Redis instance")


def _get_redis_raw(key):
    if config_store is not None:
        try:
            return config_store.get(key)
        except redis.ConnectionError:
            pass
    return None


def _get(key, fallback):
    if config_store is not None:
        try:
            value = config_store.get(key)
            if value is not None:
                return value
        except redis.ConnectionError:
            pass
    return fallback


def _sync(key, env_value):
    if config_store is not None:
        try:
            current = config_store.get(key)
            if current is None or str(current) != str(env_value):
                config_store.set(key, env_value)
        except redis.ConnectionError:
            pass
    return env_value


def get_user_limit():
    return int(_get(USER_LIMIT_KEY, config.USER_LIMIT))


def get_default_timezone():
    return _get(DEFAULT_TIMEZONE_KEY, config.DEFAULT_TIMEZONE)


def get_default_locale():
    return _get(DEFAULT_LOCALE_KEY, config.DEFAULT_LOCALE)


def get_nomad_host():
    return _get(NOMAD_HOST_KEY, config.JOB_QUEUE_NOMAD_HOST)


def get_nomad_normalize_job():
    return _get(NOMAD_NORMALIZE_JOB_KEY, config.JOB_QUEUE_NOMAD_NORMALIZE_JOB)


def get_nomad_playlist_job():
    return _get(NOMAD_PLAYLIST_JOB_KEY, config.JOB_QUEUE_NOMAD_PLAYLIST_JOB)


def get_config_comparison():
    return {
        "user_limit": {
            "env": config.USER_LIMIT,
            "redis": _get_redis_raw(USER_LIMIT_KEY),
        },
        "default_timezone": {
            "env": config.DEFAULT_TIMEZONE,
            "redis": _get_redis_raw(DEFAULT_TIMEZONE_KEY),
        },
        "default_locale": {
            "env": config.DEFAULT_LOCALE,
            "redis": _get_redis_raw(DEFAULT_LOCALE_KEY),
        },
        "nomad_host": {
            "env": config.JOB_QUEUE_NOMAD_HOST,
            "redis": _get_redis_raw(NOMAD_HOST_KEY),
        },
        "nomad_normalize_job": {
            "env": config.JOB_QUEUE_NOMAD_NORMALIZE_JOB,
            "redis": _get_redis_raw(NOMAD_NORMALIZE_JOB_KEY),
        },
        "nomad_playlist_job": {
            "env": config.JOB_QUEUE_NOMAD_PLAYLIST_JOB,
            "redis": _get_redis_raw(NOMAD_PLAYLIST_JOB_KEY),
        },
    }


def sync_config():
    """
    Read config values from environment variables and push them to
    Redis if absent or different. Called at app startup and by the
    ``reload-config`` CLI command.
    Returns a dict of synced values.
    """
    return {
        "user_limit": _sync(USER_LIMIT_KEY, config.USER_LIMIT),
        "default_timezone": _sync(
            DEFAULT_TIMEZONE_KEY, config.DEFAULT_TIMEZONE
        ),
        "default_locale": _sync(DEFAULT_LOCALE_KEY, config.DEFAULT_LOCALE),
        "nomad_host": _sync(NOMAD_HOST_KEY, config.JOB_QUEUE_NOMAD_HOST),
        "nomad_normalize_job": _sync(
            NOMAD_NORMALIZE_JOB_KEY,
            config.JOB_QUEUE_NOMAD_NORMALIZE_JOB,
        ),
        "nomad_playlist_job": _sync(
            NOMAD_PLAYLIST_JOB_KEY,
            config.JOB_QUEUE_NOMAD_PLAYLIST_JOB,
        ),
    }



from sqlalchemy.exc import IntegrityError

from flask import request
from flask_restful import Resource
from flask_jwt_extended import jwt_required

from zou.app.models.entity import Entity, EntityLink
from zou.app.models.project import Project
from zou.app.models.task import Task
from zou.app.mixin import ArgsMixin
from zou.app.utils import events, fields, permissions
from zou.app.services.exception import WrongParameterException
from zou.app.services import (
    entities_service,
    shots_service,
    tasks_service,
    user_service,
)


class BaseImportKitsuResource(Resource, ArgsMixin):
    def __init__(self, model):
        Resource.__init__(self)
        self.model = model

    @jwt_required()
    def post(self):
        """
        Import kitsu resource
        ---
        tags:
          - Import
        description: Import Kitsu resources. Send a list of Kitsu entries in
          the JSON body. Returns created or updated resources.
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      example: Resource name
                    project_id:
                      type: string
                      format: uuid
                      example: b24a6ea4-ce75-4665-a070-57453082c25
              example:
                - id: a24a6ea4-ce75-4665-a070-57453082c25
                  name: Example resource
                  project_id: b24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
              description: Resources imported successfully
              content:
                application/json:
                  schema:
                    type: array
                    items:
                      type: object
                      properties:
                        id:
                          type: string
                          format: uuid
                          example: a24a6ea4-ce75-4665-a070-57453082c25
                        name:
                          type: string
                          example: Imported resource
                        created_at:
                          type: string
                          format: date-time
                          example: "2024-01-15T10:30:00Z"
                        updated_at:
                          type: string
                          format: date-time
                          example: "2024-01-15T11:00:00Z"
            400:
              description: Invalid request body or missing required fields
        """
        kitsu_entries = request.json
        if not isinstance(kitsu_entries, list):
            raise WrongParameterException("A list of entities is expected.")

        instances = []
        for entry in kitsu_entries:
            if self.check_access(entry):
                try:
                    (instance, is_updated) = self.model.create_from_import(
                        entry
                    )
                    if is_updated:
                        self.emit_event("update", entry)
                    else:
                        self.emit_event("new", entry)
                except IntegrityError as exc:
                    raise WrongParameterException(exc.orig)
                instances.append(instance)
        return fields.serialize_models(instances)

    def emit_event(self, event_type, entry):
        pass

    def check_access(self, entry):
        return permissions.has_admin_permissions()


class ImportKitsuCommentsResource(BaseImportKitsuResource):
    def __init__(self):
        BaseImportKitsuResource.__init__(self, Entity)
        user_service.check_project_manager_access()

    @jwt_required()
    def post(self):
        """
        Import kitsu comments
        ---
        description: Import Kitsu comments. Send a list of Kitsu comment
          entries in the JSON body. Returns created or updated comments
          linked to tasks.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Kitsu ID of the comment
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    object_id:
                      type: string
                      format: uuid
                      description: Task ID the comment is linked to
                      example: b24a6ea4-ce75-4665-a070-57453082c25
                    text:
                      type: string
                      description: Comment text
                      example: "This is a comment"
                    person_id:
                      type: string
                      format: uuid
                      description: Person who created the comment
                      example: c24a6ea4-ce75-4665-a070-57453082c25
              example:
                - id: a24a6ea4-ce75-4665-a070-57453082c25
                  object_id: b24a6ea4-ce75-4665-a070-57453082c25
                  text: "This is a comment"
                  person_id: c24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Comments imported successfully
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Comment unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      text:
                        type: string
                        description: Comment text
                        example: "This is a comment"
                      created_at:
                        type: string
                        format: date-time
                        description: Creation timestamp
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        description: Update timestamp
                        example: "2024-01-15T11:00:00Z"
          400:
            description: Invalid request body or missing required fields
        """
        return super().post()

    def check_access(self, entry):
        try:
            task = tasks_service.get_task(str(entry.object_id))
            project_id = task["project_id"]
            user_service.check_project_access(project_id)
        except BaseException:
            return False
        return True

    def emit_event(self, event_type, entry):
        task = tasks_service.get_task(str(entry.object_id))
        project_id = task["project_id"]
        events.emit(
            "comment:%s" % event_type,
            {"comment_id": entry.id},
            project_id=project_id,
        )


class ImportKitsuEntitiesResource(BaseImportKitsuResource):
    def __init__(self):
        BaseImportKitsuResource.__init__(self, Entity)

    @jwt_required()
    def post(self):
        """
        Import kitsu entities
        ---
        description: Import Kitsu entities (assets, shots, sequences, etc.).
          Send a list of Kitsu entity entries in the JSON body. Returns
          created or updated entities.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Kitsu ID of the entity
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Entity name
                      example: "Asset01"
                    project_id:
                      type: string
                      format: uuid
                      description: Project ID
                      example: b24a6ea4-ce75-4665-a070-57453082c25
                    entity_type_id:
                      type: string
                      format: uuid
                      description: Entity type ID
                      example: c24a6ea4-ce75-4665-a070-57453082c25
              example:
                - id: a24a6ea4-ce75-4665-a070-57453082c25
                  name: "Asset01"
                  project_id: b24a6ea4-ce75-4665-a070-57453082c25
                  entity_type_id: c24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Entities imported successfully
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Entity unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Entity name
                        example: "Asset01"
                      created_at:
                        type: string
                        format: date-time
                        description: Creation timestamp
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        description: Update timestamp
                        example: "2024-01-15T11:00:00Z"
          400:
            description: Invalid request body or missing required fields
        """
        return super().post()

    def check_access(self, entry):
        try:
            project_id = entry["project_id"]
            user_service.check_project_access(project_id)
        except BaseException:
            return False
        return True

    def emit_event(self, event_type, entry):
        project_id = entry["project_id"]
        name = shots_service.get_base_entity_type_name(entry)
        events.emit(
            "%s:%s" % (name.lower(), event_type),
            {"%s_id" % name: entry["id"]},
            project_id=project_id,
        )


class ImportKitsuProjectsResource(BaseImportKitsuResource):
    def __init__(self):
        BaseImportKitsuResource.__init__(self, Project)

    @jwt_required()
    def post(self):
        """
        Import kitsu projects
        ---
        description: Import Kitsu projects. Send a list of Kitsu project
          entries in the JSON body. Returns created or updated projects.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Kitsu ID of the project
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Project name
                      example: "My Project"
                    production_type:
                      type: string
                      description: Production type
                      example: "tvshow"
              example:
                - id: a24a6ea4-ce75-4665-a070-57453082c25
                  name: "My Project"
                  production_type: "tvshow"
        responses:
          200:
            description: Projects imported successfully
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Project unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Project name
                        example: "My Project"
                      created_at:
                        type: string
                        format: date-time
                        description: Creation timestamp
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        description: Update timestamp
                        example: "2024-01-15T11:00:00Z"
          400:
            description: Invalid request body or missing required fields
        """
        return super().post()

    def emit_event(self, event_type, entry):
        events.emit("project:%s" % event_type, project_id=entry["id"])


class ImportKitsuTasksResource(BaseImportKitsuResource):
    def __init__(self):
        BaseImportKitsuResource.__init__(self, Task)

    @jwt_required()
    def post(self):
        """
        Import kitsu tasks
        ---
        description: Import Kitsu tasks. Send a list of Kitsu task entries in
          the JSON body. Returns created or updated tasks.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Kitsu ID of the task
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    name:
                      type: string
                      description: Task name
                      example: "Modeling"
                    project_id:
                      type: string
                      format: uuid
                      description: Project ID
                      example: b24a6ea4-ce75-4665-a070-57453082c25
                    entity_id:
                      type: string
                      format: uuid
                      description: Entity ID the task is linked to
                      example: c24a6ea4-ce75-4665-a070-57453082c25
                    task_type_id:
                      type: string
                      format: uuid
                      description: Task type ID
                      example: d24a6ea4-ce75-4665-a070-57453082c25
              example:
                - id: a24a6ea4-ce75-4665-a070-57453082c25
                  name: "Modeling"
                  project_id: b24a6ea4-ce75-4665-a070-57453082c25
                  entity_id: c24a6ea4-ce75-4665-a070-57453082c25
                  task_type_id: d24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Tasks imported successfully
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Task unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        description: Task name
                        example: "Modeling"
                      created_at:
                        type: string
                        format: date-time
                        description: Creation timestamp
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        description: Update timestamp
                        example: "2024-01-15T11:00:00Z"
          400:
            description: Invalid request body or missing required fields
        """
        return super().post()

    def check_access(self, entry):
        try:
            project_id = entry["project_id"]
            user_service.check_project_access(project_id)
        except BaseException:
            return False
        return True

    def emit_event(self, event_type, entry):
        events.emit("task:%s" % event_type, project_id=entry["project_id"])


class ImportKitsuEntityLinksResource(BaseImportKitsuResource):
    def __init__(self):
        BaseImportKitsuResource.__init__(self, EntityLink)

    @jwt_required()
    def post(self):
        """
        Import kitsu entity links
        ---
        description: Import Kitsu entity links (casting links). Send a list
          of Kitsu entity link entries in the JSON body. Returns created or
          updated entity links.
        tags:
          - Import
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Kitsu ID of the entity link
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    entity_in_id:
                      type: string
                      format: uuid
                      description: Source entity ID
                      example: b24a6ea4-ce75-4665-a070-57453082c25
                    entity_out_id:
                      type: string
                      format: uuid
                      description: Target entity ID
                      example: c24a6ea4-ce75-4665-a070-57453082c25
              example:
                - id: a24a6ea4-ce75-4665-a070-57453082c25
                  entity_in_id: b24a6ea4-ce75-4665-a070-57453082c25
                  entity_out_id: c24a6ea4-ce75-4665-a070-57453082c25
        responses:
          200:
            description: Entity links imported successfully
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Entity link unique identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      entity_in_id:
                        type: string
                        format: uuid
                        description: Source entity ID
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      entity_out_id:
                        type: string
                        format: uuid
                        description: Target entity ID
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      created_at:
                        type: string
                        format: date-time
                        description: Creation timestamp
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        description: Update timestamp
                        example: "2024-01-15T11:00:00Z"
          400:
            description: Invalid request body or missing required fields
        """
        return super().post()

    def check_access(self, entry):
        try:
            entity = entities_service.get_entity(entry["entity_in_id"])
            project_id = entity["project_id"]
            user_service.check_project_access(project_id)
        except BaseException:
            return False
        return True

    def emit_event(self, event_type, entry):
        entity = entities_service.get_entity(entry["entity_in_id"])
        project_id = entity["project_id"]
        events.emit("entity-link:%s" % event_type, project_id=project_id)



import hmac
import secrets
import urllib.parse

from flask import request, jsonify, current_app, redirect, make_response
from flask_restful import Resource
from flask_principal import (
    Identity,
    AnonymousIdentity,
    identity_changed,
)
from flask_jwt_extended import (
    jwt_required,
    create_access_token,
    create_refresh_token,
    set_access_cookies,
    set_refresh_cookies,
    unset_jwt_cookies,
    unset_refresh_cookies,
    get_jwt,
)

from sqlalchemy.exc import OperationalError, TimeoutError
from babel.dates import format_datetime
from saml2 import entity, client_base

from zou.app import app, config
from zou.app.mixin import ArgsMixin
from zou.app.utils import auth, emails, permissions, date_helpers
from zou.app.utils.email_i18n import get_email_translation
from zou.app.services import (
    persons_service,
    auth_service,
    events_service,
    templates_service,
)

from zou.app.utils.flask import is_from_browser
from zou.app.utils.saml import saml_client_for

from zou.app.stores import auth_tokens_store
from zou.app.services.exception import (
    EmailOTPAlreadyEnabledException,
    EmailOTPNotEnabledException,
    FIDONoPreregistrationException,
    FIDONotEnabledException,
    FIDOServerException,
    MissingOTPException,
    NoAuthStrategyConfigured,
    NoTwoFactorAuthenticationEnabled,
    PersonNotFoundException,
    TooMuchLoginFailedAttemps,
    TOTPAlreadyEnabledException,
    TOTPNotEnabledException,
    UnactiveUserException,
    UserCantConnectDueToNoFallback,
    WrongOTPException,
    WrongPasswordException,
    WrongUserException,
)


def _build_2fa_registration_response(response_data, user_id):
    """
    After a successful 2FA registration, re-emit JWT cookies without
    the requires_2fa_setup claim so the user gets full access.
    """
    additional_claims = {"identity_type": "person"}
    access_token = create_access_token(
        identity=user_id,
        additional_claims=additional_claims,
    )
    refresh_token = create_refresh_token(
        identity=user_id,
        additional_claims=additional_claims,
    )
    response_data["access_token"] = access_token
    response_data["refresh_token"] = refresh_token
    response = jsonify(response_data)
    if is_from_browser(request.user_agent):
        set_access_cookies(response, access_token)
        set_refresh_cookies(response, refresh_token)

    current_app.logger.info(
        "2FA setup completed, JWT refreshed for user %s." % user_id
    )
    return response


class AuthenticatedResource(Resource):

    @jwt_required()
    def get(self):
        """
        Check authentication status
        ---
        description: Returns information if the user is authenticated.
          It can be used by third party tools, especially browser frontend,
          to know if current user is still logged in.
        tags:
            - Authentication
        responses:
          200:
            description: User authenticated
          401:
            description: Person not found
        """
        person = persons_service.get_current_user(relations=True)
        person["fido_devices"] = (
            persons_service.get_current_user_fido_devices()
        )
        organisation = persons_service.get_organisation(
            sensitive=permissions.has_admin_permissions()
        )
        return {
            "authenticated": True,
            "user": person,
            "organisation": organisation,
        }


class LogoutResource(Resource):

    @jwt_required()
    @permissions.require_person
    def get(self):
        """
        Logout user
        ---
        description: Log user out by revoking auth tokens. Once logged out, current user cannot access the API anymore.
        tags:
            - Authentication
        responses:
          200:
            description: Logout successful
        """
        try:
            auth_service.logout(get_jwt()["jti"])
            identity_changed.send(
                current_app._get_current_object(), identity=AnonymousIdentity()
            )
        except KeyError:
            return {"Access token not found."}, 500

        logout_data = {"logout": True}

        if is_from_browser(request.user_agent):
            response = jsonify(logout_data)
            unset_jwt_cookies(response)
            return response
        else:
            return logout_data


class LoginResource(Resource, ArgsMixin):

    def post(self):
        """
        Login user
        ---
        description: Log in user by creating and registering auth tokens.
          Login is based on email and password. If no user matches given email
          It fallbacks to a desktop ID. It is useful for desktop tools that
          don't know user email.
          It is also possible to login with TOTP, Email OTP, FIDO and recovery
          code.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  email:
                    type: string
                    format: email
                    example: admin@example.com
                    description: User email address
                  password:
                    type: string
                    format: password
                    example: "********"
                    description: User password
                    required: true
                  totp:
                    type: string
                    example: 123456
                    description: TOTP verification code for two-factor authentication
                    required: false
                  email_otp:
                    type: string
                    example: 123456
                    description: Email OTP verification code for two-factor authentication
                  fido_authentication_response:
                    type: object
                    description: FIDO authentication response for WebAuth
                  recovery_code:
                    type: string
                    example: ABCD-EFGH-IJKL-MNOP
                    description: Recovery code for two-factor authentication
                required:
                  - email
                  - password
        responses:
          200:
            description: Login successful
          400:
            description: Login failed
        """
        (
            email,
            password,
            totp,
            email_otp,
            fido_authentication_response,
            recovery_code,
        ) = self.get_arguments()
        try:
            user = auth_service.check_auth(
                app,
                email,
                password,
                totp,
                email_otp,
                fido_authentication_response,
                recovery_code,
            )

            if auth_service.is_default_password(app, password):
                token = auth_service.generate_reset_token()
                auth_tokens_store.add(
                    "reset-token-%s" % email, token, ttl=3600 * 2
                )
                current_app.logger.info(
                    "User %s must change his password." % email
                )
                return (
                    {
                        "login": False,
                        "default_password": True,
                        "token": token,
                    },
                    400,
                )

            # Check if 2FA enforcement requires restricted access
            requires_2fa_setup = False
            if app.config["ENFORCE_2FA"]:
                if not auth_service.is_user_exempt_from_2fa(user, app):
                    if not auth_service.person_two_factor_authentication_enabled(
                        user
                    ):
                        requires_2fa_setup = True

            additional_claims = {"identity_type": "person"}
            if requires_2fa_setup:
                additional_claims["requires_2fa_setup"] = True

            access_token = create_access_token(
                identity=user["id"],
                additional_claims=additional_claims,
            )
            refresh_token = create_refresh_token(
                identity=user["id"],
                additional_claims=additional_claims,
            )
            identity_changed.send(
                current_app._get_current_object(),
                identity=Identity(user["id"], "person"),
            )

            ip_address = request.environ.get(
                "HTTP_X_REAL_IP", request.remote_addr
            )

            organisation = persons_service.get_organisation(
                sensitive=user["role"] == "admin"
            )

            response_data = {
                "user": user,
                "organisation": organisation,
                "login": True,
                "access_token": access_token,
                "refresh_token": refresh_token,
            }
            if requires_2fa_setup:
                response_data["two_factor_authentication_required"] = True

            response = jsonify(response_data)

            if is_from_browser(request.user_agent):
                set_access_cookies(response, access_token)
                set_refresh_cookies(response, refresh_token)
                events_service.create_login_log(user["id"], ip_address, "web")
            else:
                events_service.create_login_log(
                    user["id"], ip_address, "script"
                )
            if requires_2fa_setup:
                current_app.logger.info(
                    f"User {email} logged in with restricted"
                    " access - 2FA setup required."
                )
            else:
                current_app.logger.info(f"User {email} is logged in.")
            return response
        except WrongUserException:
            current_app.logger.info(f"User {email} is not registered.")
            return {"login": False}, 400
        except WrongPasswordException:
            current_app.logger.info(f"User {email} gave a wrong password.")
            return {"login": False}, 400
        except NoAuthStrategyConfigured:
            current_app.logger.info(
                "Authentication strategy is not properly configured."
            )
            return {"login": False}, 409
        except UserCantConnectDueToNoFallback:
            current_app.logger.info(
                f"User {email} can't login due to no fallback from LDAP."
            )
            return {"login": False}, 400
        except TimeoutError:
            current_app.logger.info("Timeout occurs while logging in.")
            return {"login": False}, 400
        except UnactiveUserException:
            current_app.logger.info(f"User {email} is unactive.")
            return (
                {
                    "error": True,
                    "login": False,
                    "message": "User is unactive, he cannot log in.",
                },
                401,
            )
        except TooMuchLoginFailedAttemps:
            current_app.logger.info(
                f"User {email} can't log in due to too much login failed attemps."
            )
            return (
                {
                    "error": True,
                    "login": False,
                    "too_many_failed_login_attemps": True,
                },
                400,
            )
        except MissingOTPException as e:
            current_app.logger.info(
                f"User {email} can't log in due to missing OTP."
            )
            return (
                {
                    "error": True,
                    "login": False,
                    "missing_OTP": True,
                    "preferred_two_factor_authentication": e.preferred_two_factor_authentication,
                    "two_factor_authentication_enabled": e.two_factor_authentication_enabled,
                },
                400,
            )
        except WrongOTPException:
            current_app.logger.info(
                f"User {email} can't log in due to wrong OTP."
            )
            return (
                {
                    "error": True,
                    "login": False,
                    "wrong_OTP": True,
                },
                400,
            )
        except OperationalError as exception:
            current_app.logger.error(exception, exc_info=1)
            return (
                {
                    "error": True,
                    "login": False,
                    "message": "Database doesn't seem reachable.",
                },
                500,
            )
        except Exception as exception:
            current_app.logger.error(exception, exc_info=1)
            return {
                "error": True,
                "login": False,
                "message": "A server error occurred. Please contact your administrator.",
            }, 500

    def get_arguments(self):
        args = self.get_args(
            [
                {
                    "name": "email",
                    "required": True,
                    "help": "User email is missing.",
                },
                ("password", "default"),
                "totp",
                "email_otp",
                ("fido_authentication_response", None, False, dict),
                "recovery_code",
            ]
        )

        return (
            args["email"],
            args["password"],
            args["totp"],
            args["email_otp"],
            args["fido_authentication_response"],
            args["recovery_code"],
        )


class RefreshTokenResource(Resource):
    @jwt_required(refresh=True)
    @permissions.require_person
    def get(self):
        """
        Refresh access token
        ---
        description: Tokens are considered outdated every two weeks.
          This route allows to extend their lifetime before they get outdated.
        tags:
            - Authentication
        responses:
          200:
            description: Access Token
        """
        user = persons_service.get_current_user()
        additional_claims = {"identity_type": "person"}

        if app.config["ENFORCE_2FA"]:
            user_unsafe = persons_service.get_current_user(unsafe=True)
            if not auth_service.is_user_exempt_from_2fa(user_unsafe, app):
                if not auth_service.person_two_factor_authentication_enabled(
                    user_unsafe
                ):
                    additional_claims["requires_2fa_setup"] = True

        access_token = create_access_token(
            identity=user["id"],
            additional_claims=additional_claims,
        )
        if is_from_browser(request.user_agent):
            response = jsonify({"refresh": True})
            set_access_cookies(response, access_token)
            unset_refresh_cookies(response)
            return response
        else:
            return {"access_token": access_token}


class RegistrationResource(Resource, ArgsMixin):
    """
    Allow a user to register himself to the service.
    """

    def post(self):
        """
        Register new user
        ---
        description: Allow a user to register himself to the service.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  email:
                    type: string
                    format: email
                    example: admin@example.com
                    description: User email address
                  password:
                    type: string
                    format: password
                    description: User password
                  password_2:
                    type: string
                    format: password
                    description: Password confirmation
                  first_name:
                    type: string
                    description: User first name
                  last_name:
                    type: string
                    description: User last name
                required:
                  - email
                  - password
                  - password_2
                  - first_name
                  - last_name
        responses:
          201:
            description: Registration successful
          400:
            description: Invalid password or email
        """
        (
            email,
            password,
            password_2,
            first_name,
            last_name,
        ) = self.get_arguments()

        try:
            email = auth.validate_email(email)
            auth.validate_password(password, password_2)
            password = auth.encrypt_password(password)
            persons_service.create_person(
                email, password, first_name, last_name
            )
            return {"registration_success": True}, 201
        except auth.PasswordsNoMatchException:
            return (
                {
                    "error": True,
                    "message": "Confirmation password doesn't match.",
                },
                400,
            )
        except auth.PasswordTooShortException:
            return {"error": True, "message": "Password is too short."}, 400
        except auth.EmailNotValidException as exception:
            return {"error": True, "message": str(exception)}, 400

    def get_arguments(self):
        args = self.get_args(
            [
                {
                    "name": "email",
                    "required": True,
                    "help": "User email is missing.",
                },
                {
                    "name": "first_name",
                    "required": True,
                    "help": "First name is missing.",
                },
                {
                    "name": "last_name",
                    "required": True,
                    "help": "Last name is missing.",
                },
                {
                    "name": "password",
                    "required": True,
                    "help": "Password is missing.",
                },
                {
                    "name": "password_2",
                    "required": True,
                    "help": "Confirmation password is missing.",
                },
            ]
        )

        return (
            args["email"],
            args["password"],
            args["password_2"],
            args["first_name"],
            args["last_name"],
        )


class ChangePasswordResource(Resource, ArgsMixin):

    @jwt_required()
    @permissions.require_person
    def post(self):
        """
        Change user password
        ---
        description: Allow the user to change his password. Requires current
          password for verification and password confirmation to ensure
          accuracy.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  old_password:
                    type: string
                    format: password
                    description: Current password
                  password:
                    type: string
                    format: password
                    description: New password
                  password_2:
                    type: string
                    format: password
                    description: New password confirmation
                required:
                  - old_password
                  - password
                  - password_2
        responses:
          200:
            description: Password changed
          400:
            description: Invalid password or inactive user
        """
        old_password, password, password_2 = self.get_arguments()

        try:
            user = persons_service.get_current_user()
            auth_service.check_auth(
                app, user["email"], old_password, no_otp=True
            )
            auth.validate_password(password, password_2)
            password = auth.encrypt_password(password)
            persons_service.update_password(user["email"], password)
            current_app.logger.info(
                "User %s has changed his password" % user["email"]
            )
            organisation = persons_service.get_organisation()
            locale = user.get("locale") or getattr(
                config, "DEFAULT_LOCALE", "en_US"
            )
            if hasattr(locale, "language"):
                locale = str(locale)
            time_string = format_datetime(
                date_helpers.get_utc_now_datetime(),
                tzinfo=user["timezone"],
                locale=user["locale"],
            )
            person_IP = request.headers.get("X-Forwarded-For", None) or ""
            subject = get_email_translation(
                locale,
                "auth_password_changed_subject",
                organisation_name=organisation["name"],
            )
            title = get_email_translation(
                locale, "auth_password_changed_title"
            )
            html = get_email_translation(
                locale,
                "auth_password_changed_body",
                first_name=user["first_name"],
                time_string=time_string,
                person_IP=person_IP,
            )
            email_html_body = templates_service.generate_html_body(
                title, html, locale=locale
            )
            emails.send_email(
                subject, email_html_body, user["email"], locale=locale
            )
            return {"success": True}

        except auth.PasswordsNoMatchException:
            return (
                {
                    "error": True,
                    "message": "Confirmation password doesn't match.",
                },
                400,
            )
        except auth.PasswordTooShortException:
            return {"error": True, "message": "Password is too short."}, 400
        except UnactiveUserException:
            return {"error": True, "message": "User is unactive."}, 400
        except WrongPasswordException:
            return {"error": True, "message": "Old password is wrong."}, 400

    def get_arguments(self):
        args = self.get_args(
            [
                {
                    "name": "old_password",
                    "required": True,
                    "help": "Old password is missing.",
                },
                {
                    "name": "password",
                    "required": True,
                    "help": "New password is missing.",
                },
                {
                    "name": "password_2",
                    "required": True,
                    "help": "New password confirmation is missing.",
                },
            ]
        )

        return (args["old_password"], args["password"], args["password_2"])


class ResetPasswordResource(Resource, ArgsMixin):

    def put(self):
        """
        Reset password with token
        ---
        description: Allow a user to change his password when he forgets it.
          It uses a token sent by email to the user to verify it is the user
          who requested the password reset.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  email:
                    type: string
                    format: email
                    example: admin@example.com
                    description: User email address
                  token:
                    type: string
                    format: JWT token
                    description: Password reset token
                  password:
                    type: string
                    format: password
                    description: New password
                  password2:
                    type: string
                    format: password
                    description: New password confirmation
                required:
                  - email
                  - token
                  - password
                  - password2
        responses:
          200:
            description: Password reset
          400:
            description: Invalid password
                         Wrong or expired token
                         Inactive user
        """
        args = self.get_args(
            [
                ("email", "", True),
                ("token", "", True),
                ("password", "", True),
                ("password2", "", True),
            ]
        )

        try:
            token_from_store = auth_tokens_store.get(
                "reset-token-%s" % args["email"]
            )
            if token_from_store and hmac.compare_digest(
                token_from_store, args["token"]
            ):
                auth.validate_password(args["password"], args["password2"])
                password = auth.encrypt_password(args["password"])
                persons_service.update_password(args["email"], password)
                auth_tokens_store.delete("reset-token-%s" % args["email"])
                current_app.logger.info(
                    "User %s has reset his password" % args["email"]
                )
                return {"success": True}
            else:
                return (
                    {"error": True, "message": "Wrong or expired token."},
                    400,
                )

        except auth.PasswordsNoMatchException:
            return (
                {
                    "error": True,
                    "message": "Confirmation password doesn't match.",
                },
                400,
            )
        except auth.PasswordTooShortException:
            return {"error": True, "message": "Password is too short."}, 400
        except UnactiveUserException:
            return {"error": True, "message": "User is inactive."}, 400

    def post(self):
        """
        Request password reset
        ---
        description: Send a password reset token by email to the user.
          It uses a classic scheme where a token is sent by email.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  email:
                    type: string
                    format: email
                    example: admin@example.com
                    description: User email address
                required:
                  - email
        responses:
          200:
            description: Reset token sent
          400:
            description: Email not listed in database
        """
        args = self.get_args([("email", "", True)])

        try:
            user = persons_service.get_person_by_email(args["email"])
            if not user["active"]:
                return (
                    {"error": True, "message": "This user is inactive."},
                    400,
                )
        except PersonNotFoundException:
            return (
                {"error": True, "message": "Email not listed in database."},
                400,
            )

        token = auth_service.generate_reset_token()
        auth_tokens_store.add(
            "reset-token-%s" % args["email"], token, ttl=3600 * 2
        )
        params = {"email": args["email"], "token": token}
        query = urllib.parse.urlencode(params)
        reset_url = "%s://%s/reset-change-password?%s" % (
            config.DOMAIN_PROTOCOL,
            config.DOMAIN_NAME,
            query,
        )
        locale = user.get("locale") or getattr(
            config, "DEFAULT_LOCALE", "en_US"
        )
        if hasattr(locale, "language"):
            locale = str(locale)
        time_string = format_datetime(
            date_helpers.get_utc_now_datetime(),
            tzinfo=user["timezone"],
            locale=user["locale"],
        )
        person_IP = request.headers.get("X-Forwarded-For", None) or ""
        organisation = persons_service.get_organisation()
        subject = get_email_translation(
            locale,
            "auth_password_recovery_subject",
            organisation_name=organisation["name"],
        )
        title = get_email_translation(locale, "auth_password_recovery_title")
        html = get_email_translation(
            locale,
            "auth_password_recovery_body",
            first_name=user["first_name"],
            reset_url=reset_url,
            time_string=time_string,
            person_IP=person_IP,
        )
        email_html_body = templates_service.generate_html_body(
            title, html, locale=locale
        )
        emails.send_email(
            subject, email_html_body, args["email"], locale=locale
        )
        return {"success": "Reset token sent"}


class TOTPResource(Resource, ArgsMixin):

    @jwt_required()
    @permissions.require_person
    def put(self):
        """
        Pre-enable TOTP
        ---
        description: Prepare TOTP (Time-based One-Time Password) for enabling.
          It returns provisioning URI and secret for authenticator app setup.
        tags:
            - Authentication
        responses:
          200:
            description: TOTP pre-enabled
          400:
            description: TOTP already enabled
        """
        try:
            totp_provisionning_uri, totp_secret = auth_service.pre_enable_totp(
                persons_service.get_current_user()["id"]
            )
            return {
                "totp_provisionning_uri": totp_provisionning_uri,
                "otp_secret": totp_secret,
            }
        except TOTPAlreadyEnabledException:
            return (
                {"error": True, "message": "TOTP already enabled."},
                400,
            )

    @jwt_required()
    @permissions.require_person
    def post(self):
        """
        Enable TOTP
        ---
        description: Enable TOTP (Time-based One-Time Password) authentication.
          It requires verification code from authenticator app.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  totp:
                    type: string
                    description: TOTP verification code from authenticator app
                required:
                  - totp
        responses:
          200:
            description: TOTP enabled
          400:
            description: TOTP already enabled or verification failed
        """
        args = self.get_args([("totp", "", True)])

        try:
            current_user = persons_service.get_current_user()
            otp_recovery_codes = auth_service.enable_totp(
                current_user["id"], args["totp"]
            )
            return _build_2fa_registration_response(
                {"otp_recovery_codes": otp_recovery_codes},
                current_user["id"],
            )
        except TOTPAlreadyEnabledException:
            return (
                {"error": True, "message": "TOTP already enabled."},
                400,
            )
        except WrongOTPException:
            return (
                {
                    "error": True,
                    "message": "TOTP verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )

    @jwt_required()
    @permissions.require_person
    def delete(self):
        """
        Disable TOTP
        ---
        description: Disable TOTP (Time-based One-Time Password) authentication.
          It requires two-factor authentication verification.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  totp:
                    type: string
                    description: TOTP verification code
                  email_otp:
                    type: string
                    description: Email OTP verification code
                  fido_authentication_response:
                    type: object
                    description: FIDO authentication response
                  recovery_code:
                    type: string
                    description: Recovery code for two-factor authentication
        responses:
          200:
            description: TOTP disabled
          400:
            description: TOTP not enabled or verification failed
        """
        args = self.get_args(
            [
                ("totp", None, False),
                ("email_otp", None, False),
                ("fido_authentication_response", {}, False, dict),
                ("recovery_code", None, False),
            ]
        )

        try:
            person = persons_service.get_current_user(unsafe=True)
            if not auth_service.person_two_factor_authentication_enabled(
                person
            ):
                raise TOTPNotEnabledException
            if not auth_service.check_two_factor_authentication(
                person,
                args["totp"],
                args["email_otp"],
                args["fido_authentication_response"],
                args["recovery_code"],
            ):
                raise WrongOTPException
            auth_service.disable_totp(person["id"])
            return {"success": True}
        except TOTPNotEnabledException:
            return (
                {"error": True, "message": "TOTP not enabled."},
                400,
            )
        except (WrongOTPException, MissingOTPException):
            return (
                {
                    "error": True,
                    "message": "OTP verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )


class EmailOTPResource(Resource, ArgsMixin):

    def get(self):
        """
        Send email OTP
        ---
        description: Send a one-time password by email to the user for
          authentication.
        tags:
            - Authentication
        parameters:
          - in: query
            name: email
            required: True
            type: string
            format: email
            description: User email address
        responses:
          200:
            description: OTP by email sent
          400:
            description: OTP by email not enabled
        """
        args = self.get_args(
            [
                ("email", None, True),
            ],
            location="values",
        )

        try:
            try:
                person = persons_service.get_person_by_email_desktop_login(
                    args["email"]
                )
            except PersonNotFoundException:
                raise WrongUserException()
            if not person["email_otp_enabled"]:
                raise EmailOTPNotEnabledException
            auth_service.send_email_otp(person)
            return {"success": True}
        except EmailOTPNotEnabledException:
            return (
                {"error": True, "message": "OTP by email not enabled."},
                400,
            )
        except WrongUserException:
            return (
                {"error": True, "message": "User not found."},
                404,
            )

    @jwt_required()
    @permissions.require_person
    def put(self):
        """
        Pre-enable email OTP
        ---
        description: Prepare email OTP (One-Time Password) for enabling.
          It sets up email-based two-factor authentication.
        tags:
            - Authentication
        responses:
          200:
            description: Email OTP pre-enabled
          400:
            description: Email OTP already enabled
        """
        try:
            auth_service.pre_enable_email_otp(
                persons_service.get_current_user()["id"]
            )
            return {"success": True}
        except EmailOTPAlreadyEnabledException:
            return (
                {"error": True, "message": "OTP by email already enabled."},
                400,
            )

    @jwt_required()
    @permissions.require_person
    def post(self):
        """
        Enable email OTP
        ---
        description: Enable email OTP (One-Time Password) authentication.
          It requires verification code sent to email.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  email_otp:
                    type: string
                    description: Email OTP verification code
                required:
                  - email_otp
        responses:
          200:
            description: Email OTP enabled
          400:
            description: Email OTP already enabled or verification failed
        """
        args = self.get_args([("email_otp", "", True)])

        try:
            current_user = persons_service.get_current_user()
            otp_recovery_codes = auth_service.enable_email_otp(
                current_user["id"],
                args["email_otp"],
            )
            return _build_2fa_registration_response(
                {"otp_recovery_codes": otp_recovery_codes},
                current_user["id"],
            )
        except EmailOTPAlreadyEnabledException:
            return (
                {"error": True, "message": "OTP by email already enabled."},
                400,
            )
        except WrongOTPException:
            return (
                {
                    "error": True,
                    "message": "OTP by email verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )

    @jwt_required()
    @permissions.require_person
    def delete(self):
        """
        Disable email OTP
        ---
        description: Disable email OTP (One-Time Password) authentication.
          It requires two-factor authentication verification.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  totp:
                    type: string
                    description: TOTP verification code
                  email_otp:
                    type: string
                    description: Email OTP verification code
                  fido_authentication_response:
                    type: object
                    description: FIDO authentication response
                  recovery_code:
                    type: string
                    description: Recovery code for two-factor authentication
        responses:
          200:
            description: Email OTP disabled
          400:
            description: Email OTP not enabled or verification failed
        """
        args = self.get_args(
            [
                ("totp", None, False),
                ("email_otp", None, False),
                ("fido_authentication_response", {}, False, dict),
                ("recovery_code", None, False),
            ]
        )

        try:
            person = persons_service.get_current_user(unsafe=True)
            if not auth_service.person_two_factor_authentication_enabled(
                person
            ):
                raise EmailOTPNotEnabledException
            if not auth_service.check_two_factor_authentication(
                person,
                args["totp"],
                args["email_otp"],
                args["fido_authentication_response"],
                args["recovery_code"],
            ):
                raise WrongOTPException
            auth_service.disable_email_otp(person["id"])
            return {"success": True}
        except EmailOTPNotEnabledException:
            return (
                {"error": True, "message": "OTP by email not enabled."},
                400,
            )
        except (WrongOTPException, MissingOTPException):
            return (
                {
                    "error": True,
                    "message": "OTP verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )


class FIDOResource(Resource, ArgsMixin):
    """
    Resource to allow a user to register/unregister FIDO device or to get a
    challenge for a FIDO device.
    """

    def get(self):
        """
        Get FIDO challenge
        ---
        description: Get a challenge for FIDO device authentication.
          It is used for WebAuthn authentication flow.
        tags:
            - Authentication
        parameters:
          - in: query
            name: email
            required: True
            type: string
            format: email
            description: User email address
        responses:
          200:
            description: FIDO challenge generated
          400:
            description: FIDO not enabled
        """
        args = self.get_args(
            [
                ("email", None, True),
            ],
            location="values",
        )

        try:
            try:
                person = persons_service.get_person_by_email_desktop_login(
                    args["email"]
                )
            except PersonNotFoundException:
                raise WrongUserException()
            if not person["fido_enabled"]:
                raise FIDONotEnabledException
            return auth_service.get_challenge_fido(person["id"])
        except FIDONotEnabledException:
            return (
                {"error": True, "message": "FIDO not enabled."},
                400,
            )
        except WrongUserException:
            return (
                {"error": True, "message": "User not found."},
                404,
            )

    @jwt_required()
    @permissions.require_person
    def put(self):
        """
        Pre-register FIDO device
        ---
        description: Prepare FIDO device for registration.
          It returns registration options for WebAuthn.
        tags:
            - Authentication
        responses:
          200:
            description: FIDO device pre-registered data
          400:
            description: Invalid request
        """
        return auth_service.pre_register_fido(
            persons_service.get_current_user()["id"]
        )

    @permissions.require_person
    @jwt_required()
    def post(self):
        """
        Register FIDO device
        ---
        description: Register a FIDO device for WebAuthn authentication.
          It requires registration response from the device.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  registration_response:
                    type: object
                    description: FIDO device registration response
                  device_name:
                    type: string
                    description: Name for the FIDO device
                required:
                  - registration_response
                  - device_name
        responses:
          200:
            description: FIDO device registered
          400:
            description: Registration failed or no preregistration
        """
        try:
            args = self.get_args(
                [
                    ("registration_response", {}, True, dict),
                    ("device_name", "", True),
                ]
            )

            current_user = persons_service.get_current_user()
            otp_recovery_codes = auth_service.register_fido(
                current_user["id"],
                args["registration_response"],
                args["device_name"],
            )
            return _build_2fa_registration_response(
                {"otp_recovery_codes": otp_recovery_codes},
                current_user["id"],
            )
        except FIDONoPreregistrationException:
            return (
                {"error": True, "message": "No preregistration before."},
                400,
            )
        except FIDOServerException:
            return (
                {
                    "error": True,
                    "message": "FIDO server exception your registration response is probly wrong.",
                },
                400,
            )

    @jwt_required()
    @permissions.require_person
    def delete(self):
        """
        Unregister FIDO device
        ---
        description: Unregister a FIDO device from WebAuthn authentication.
          The user must be authenticated.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  device_name:
                    type: string
                    description: Name of the FIDO device to unregister
                required:
                  - device_name
        responses:
          200:
            description: FIDO device unregistered
          400:
            description: FIDO not enabled
        """
        args = self.get_args(
            [
                ("device_name", None, True),
                ("totp", None, False),
                ("email_otp", None, False),
                ("fido_authentication_response", {}, False, dict),
                ("recovery_code", None, False),
            ]
        )

        try:
            person = persons_service.get_current_user(unsafe=True)
            if not person["fido_enabled"]:
                raise FIDONotEnabledException
            if not auth_service.check_two_factor_authentication(
                person,
                args["totp"],
                args["email_otp"],
                args["fido_authentication_response"],
                args["recovery_code"],
            ):
                raise WrongOTPException
            auth_service.unregister_fido(person["id"], args["device_name"])
            return {"success": True}
        except (WrongOTPException, MissingOTPException):
            return (
                {"error": True, "message": "Wrong OTP."},
                400,
            )
        except FIDONotEnabledException:
            return (
                {"error": True, "message": "FIDO not enabled."},
                400,
            )


class RecoveryCodesResource(Resource, ArgsMixin):

    @jwt_required()
    @permissions.require_person
    def put(self):
        """
        Generate recovery codes
        ---
        description: Generate new recovery codes for two-factor authentication.
          It requires two-factor authentication verification.
        tags:
            - Authentication
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  totp:
                    type: string
                    description: TOTP verification code
                  email_otp:
                    type: string
                    description: Email OTP verification code
                  fido_authentication_response:
                    type: object
                    description: FIDO authentication response
                  recovery_code:
                    type: string
                    description: Recovery code for two-factor authentication
        responses:
          200:
            description: New recovery codes generated
          400:
            description: No two-factor authentication enabled or verification failed
        """
        args = self.get_args(
            [
                ("totp", None, False),
                ("email_otp", None, False),
                ("fido_authentication_response", {}, False, dict),
                ("recovery_code", None, False),
            ]
        )

        try:
            person = persons_service.get_current_user(unsafe=True)
            if not auth_service.person_two_factor_authentication_enabled(
                person
            ):
                raise NoTwoFactorAuthenticationEnabled
            if not auth_service.check_two_factor_authentication(
                person,
                args["totp"],
                args["email_otp"],
                args["fido_authentication_response"],
                args["recovery_code"],
            ):
                raise WrongOTPException
            otp_recovery_codes = auth_service.generate_new_recovery_codes(
                person["id"]
            )
            return {"otp_recovery_codes": otp_recovery_codes}
        except WrongOTPException:
            return (
                {
                    "error": True,
                    "message": "OTP verification failed.",
                    "wrong_OTP": True,
                },
                400,
            )
        except NoTwoFactorAuthenticationEnabled:
            return (
                {
                    "error": True,
                    "message": "No two factor authentication enabled.",
                },
                400,
            )


class SAMLSSOResource(Resource, ArgsMixin):
    def post(self):
        """
        SAML SSO login
        ---
        description: Handle SAML SSO login response. Processes authentication
          response from SAML identity provider and creates a new user if they
          don't exist.
        tags:
            - Authentication
        responses:
          302:
            description: Login successful, redirect to home page
          400:
            description: SAML not enabled or wrong parameter
        """
        if not config.SAML_ENABLED:
            return {"error": "SAML is not enabled."}, 400
        authn_response = current_app.extensions[
            "saml_client"
        ].parse_authn_request_response(
            request.form["SAMLResponse"], entity.BINDING_HTTP_POST
        )
        authn_response.get_identity()
        email = authn_response.get_subject().text
        person_info = {
            k: (
                " ".join(v)
                if isinstance(v, list) and k in ["first_name", "last_name"]
                else v
            )
            for k, v in authn_response.ava.items()
            if k
            in [
                "first_name",
                "last_name",
                "phone",
                "departments",
                "studio_id",
            ]
        }
        try:
            user = persons_service.get_person_by_email(email)
            for k, v in person_info.items():
                if user.get(k) != v:
                    persons_service.update_person(
                        user["id"], person_info, bypass_protected_accounts=True
                    )
                    break
        except PersonNotFoundException:
            random_password = auth.encrypt_password(secrets.token_urlsafe(64))
            user = persons_service.create_person(
                email, random_password, **person_info
            )

        response = make_response(
            redirect(f"{config.DOMAIN_PROTOCOL}://{config.DOMAIN_NAME}")
        )

        if user["active"]:
            access_token = create_access_token(
                identity=user["id"],
                additional_claims={
                    "identity_type": "person",
                },
            )
            refresh_token = create_refresh_token(
                identity=user["id"],
                additional_claims={
                    "identity_type": "person",
                },
            )
            identity_changed.send(
                current_app._get_current_object(),
                identity=Identity(user["id"], "person"),
            )

            ip_address = request.environ.get(
                "HTTP_X_REAL_IP", request.remote_addr
            )

            set_access_cookies(response, access_token)
            set_refresh_cookies(response, refresh_token)
            events_service.create_login_log(user["id"], ip_address, "web")

        return response


class SAMLLoginResource(Resource, ArgsMixin):

    def get(self):
        """
        SAML SSO login redirect
        ---
        description: Initiate SAML SSO login by redirecting to SAML identity
          provider.
        tags:
            - Authentication
        responses:
          302:
            description: Redirect to SAML identity provider
          400:
            description: SAML not enabled or wrong parameter
        """
        if not config.SAML_ENABLED:
            return {"error": "SAML is not enabled."}, 400

        try:
            _, info = current_app.extensions[
                "saml_client"
            ].prepare_for_authenticate()
        except client_base.SAMLError:
            # retry with new client
            current_app.extensions["saml_client"] = saml_client_for(
                config.SAML_METADATA_URL
            )
            _, info = current_app.extensions[
                "saml_client"
            ].prepare_for_authenticate()

        redirect_url = None

        # Select the IdP URL to send the AuthN request to
        for key, value in info["headers"]:
            if key == "Location":
                redirect_url = value

        return redirect(redirect_url, code=302)



"""
This module is a wrapper for flask_caching. It configures it and rename
the memoize function. The aim with that cache is to minimize the requests
made on the target database.
"""

import redis

from flask_caching import Cache
from zou.app import config

cache = None

if config.CACHE_TYPE is not None:
    cache = Cache(config={"CACHE_TYPE": config.CACHE_TYPE})
else:
    try:
        redis_cache = redis.StrictRedis(
            host=config.KEY_VALUE_STORE["host"],
            port=config.KEY_VALUE_STORE["port"],
            db=config.MEMOIZE_DB_INDEX,
            password=config.KEY_VALUE_STORE["password"],
            decode_responses=True,
        )
        redis_cache.get("test")
        cache = Cache(
            config={
                "CACHE_TYPE": "redis",
                "CACHE_REDIS_HOST": config.KEY_VALUE_STORE["host"],
                "CACHE_REDIS_PORT": config.KEY_VALUE_STORE["port"],
                "CACHE_REDIS_DB": config.MEMOIZE_DB_INDEX,
                "CACHE_REDIS_PASSWORD": config.KEY_VALUE_STORE["password"],
            }
        )
    except redis.ConnectionError:
        cache = Cache(config={"CACHE_TYPE": "simple"})

memoize_function = cache.memoize


def invalidate(*args):
    cache.delete_memoized(*args)


def clear():
    cache.clear()



"""
Pydantic schemas for request body validation in the playlists blueprint.

Use these schemas with validate_request_body() to ensure incoming JSON bodies
match the expected format and to return clear validation errors (400 with
field-level messages).
"""

from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel, Field, field_validator


def _ensure_uuid(v, field_name: str):
    """Validate and coerce string to UUID for Pydantic."""
    if isinstance(v, UUID):
        return v
    if isinstance(v, str):
        from zou.app.utils.fields import is_valid_id

        if not is_valid_id(v):
            raise ValueError(f"{field_name} must be a valid UUID")
        return v
    return v


class AddEntityToPlaylistSchema(BaseModel):
    """Body for adding an entity to a playlist."""

    entity_id: UUID = Field(..., description="Entity unique identifier to add")
    preview_file_id: Optional[UUID] = Field(
        None,
        description="Optional preview file identifier associated to the entity",
    )

    @field_validator("entity_id", mode="before")
    @classmethod
    def validate_entity_id(cls, v):
        return _ensure_uuid(v, "entity_id")

    @field_validator("preview_file_id", mode="before")
    @classmethod
    def validate_preview_file_id(cls, v):
        if v is None or v == "":
            return None
        return _ensure_uuid(v, "preview_file_id")


class TempPlaylistCreateSchema(BaseModel):
    """Body for generating a temporary playlist from task IDs."""

    task_ids: List[UUID] = Field(
        ...,
        min_length=0,
        description="List of task unique identifiers",
    )

    @field_validator("task_ids", mode="before")
    @classmethod
    def validate_task_ids(cls, v):
        if not isinstance(v, list):
            raise ValueError("task_ids must be a list")
        result = []
        for i, item in enumerate(v):
            try:
                result.append(_ensure_uuid(item, f"task_ids[{i}]"))
            except ValueError as e:
                raise ValueError(f"task_ids[{i}]: invalid UUID") from e
        return result


class NotifyClientsPlaylistSchema(BaseModel):
    """Optional body for notifying clients that a playlist is ready."""

    studio_id: Optional[UUID] = Field(
        None,
        description="Studio unique identifier to notify",
    )
    department_id: Optional[UUID] = Field(
        None,
        description="Department unique identifier to notify",
    )

    @field_validator("studio_id", mode="before")
    @classmethod
    def validate_studio_id(cls, v):
        if v is None or v == "":
            return None
        return _ensure_uuid(v, "studio_id")

    @field_validator("department_id", mode="before")
    @classmethod
    def validate_department_id(cls, v):
        if v is None or v == "":
            return None
        return _ensure_uuid(v, "department_id")



import sys
from pathlib import Path

# The tests package lives at the zou repository root, not inside the zou
# package.  Add the repo root to sys.path so "tests.base" is importable
# regardless of the caller's working directory.
sys.path.insert(0, str(Path(__file__).resolve().parents[2]))

from tests.base import ApiTestCase, ApiDBTestCase  # noqa: E402

__all__ = ["ApiTestCase", "ApiDBTestCase"]



from flask_restful import Resource, inputs
from flask_jwt_extended import jwt_required

from zou.app.mixin import ArgsMixin
from zou.app.services import (
    news_service,
    projects_service,
    user_service,
    persons_service,
)
from zou.app.services.exception import NewsNotFoundException
from zou.app.utils import permissions


class NewsMixin(ArgsMixin):

    def get_news(self, project_ids=None):
        (
            only_preview,
            task_type_id,
            task_status_id,
            episode_id,
            person_id,
            page,
            limit,
            after,
            before,
        ) = self.get_arguments()

        current_user = persons_service.get_current_user_raw()

        after = self.parse_date_parameter(after)
        before = self.parse_date_parameter(before)
        result = news_service.get_last_news_for_project(
            project_ids=project_ids,
            only_preview=only_preview,
            task_type_id=task_type_id,
            task_status_id=task_status_id,
            episode_id=episode_id,
            author_id=person_id,
            page=page,
            limit=limit,
            after=after,
            before=before,
            current_user=current_user,
        )
        stats = news_service.get_news_stats_for_project(
            project_ids=project_ids,
            only_preview=only_preview,
            task_type_id=task_type_id,
            task_status_id=task_status_id,
            episode_id=episode_id,
            author_id=person_id,
            after=after,
            before=before,
            current_user=current_user,
        )
        result["stats"] = stats
        return result

    def get_arguments(self):
        args = self.get_args(
            [
                (
                    "only_preview",
                    False,
                    False,
                    inputs.boolean,
                ),
                "task_type_id",
                "task_status_id",
                "person_id",
                "project_id",
                "episode_id",
                {"name": "page", "default": 1, "type": int},
                {"name": "limit", "default": 50, "type": int},
                "after",
                "before",
            ],
        )
        return (
            args["only_preview"],
            args["task_type_id"],
            args["task_status_id"],
            args["episode_id"],
            args["person_id"],
            args["page"],
            args["limit"],
            args["after"],
            args["before"],
        )


class ProjectNewsResource(Resource, NewsMixin, ArgsMixin):

    @jwt_required()
    def get(self, project_id):
        """
        Get project latest news
        ---
        description: Get the 50 latest news object (activity feed) for a project
        tags:
          - News
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the project
          - in: query
            name: before
            required: false
            schema:
              type: string
              format: date
            example: "2022-07-12"
            description: Filter news before this date
          - in: query
            name: after
            required: false
            schema:
              type: string
              format: date
            example: "2022-07-12"
            description: Filter news after this date
          - in: query
            name: page
            required: false
            schema:
              type: integer
              default: 1
            example: 1
            description: Page number for pagination
          - in: query
            name: limit
            required: false
            schema:
              type: integer
              default: 50
            example: 50
            description: Number of news items per page
          - in: query
            name: person_id
            required: false
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter news by specific team member
          - in: query
            name: task_type_id
            required: false
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter news by task type
          - in: query
            name: task_status_id
            required: false
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter news by task status
          - in: query
            name: episode_id
            required: false
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter news by specific episode
          - in: query
            name: only_preview
            required: false
            schema:
              type: boolean
              default: false
            example: false
            description: Show only news related to preview uploads
        responses:
          '200':
            description: All news related to given project
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    data:
                      type: array
                      items:
                        type: object
                        properties:
                          id:
                            type: string
                            format: uuid
                            description: Unique news item identifier
                          title:
                            type: string
                            description: News item title
                          content:
                            type: string
                          created_at:
                            type: string
                            format: date-time
                          author_id:
                            type: string
                            format: uuid
                    stats:
                      type: object
                      properties:
                        total:
                          type: integer
          '404':
            description: Project not found
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        return self.get_news([project_id])


class NewsResource(Resource, NewsMixin, ArgsMixin):

    @jwt_required()
    def get(self):
        """
        Get open projects news
        ---
        description: Returns the latest news and activity feed from all
          projects the user has access to.
        tags:
          - News
        parameters:
          - in: query
            name: project_id
            required: false
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter news by specific project
          - in: query
            name: before
            required: false
            schema:
              type: string
              format: date
            example: "2022-07-12"
            description: Filter news before this date
          - in: query
            name: after
            required: false
            schema:
              type: string
              format: date
            example: "2022-07-12"
            description: Filter news after this date
          - in: query
            name: page
            required: false
            schema:
              type: integer
              default: 1
            example: 1
            description: Page number for pagination
          - in: query
            name: limit
            required: false
            schema:
              type: integer
              default: 50
            example: 50
            description: Number of news items per page
          - in: query
            name: person_id
            required: false
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter news by specific team member
          - in: query
            name: task_type_id
            required: false
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter news by task type
          - in: query
            name: task_status_id
            required: false
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter news by task status
          - in: query
            name: episode_id
            required: false
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Filter news by specific episode
          - in: query
            name: only_preview
            required: false
            schema:
              type: boolean
              default: false
            example: false
            description: Show only news related to preview uploads
        responses:
          '200':
            description: News feed successfully retrieved
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    data:
                      type: array
                      items:
                        type: object
                      description: Array of news items
                    stats:
                      type: object
                      description: News statistics
                    total:
                      type: integer
                      description: Total number of news items
        """
        open_project_ids = []
        if permissions.has_admin_permissions():
            open_project_ids = projects_service.open_project_ids()
        else:
            open_project_ids = user_service.get_open_project_ids()

        project_id = self.get_text_parameter("project_id")
        if project_id is not None and project_id in open_project_ids:
            open_project_ids = [project_id]

        return self.get_news(project_ids=open_project_ids)


class ProjectSingleNewsResource(Resource):

    @jwt_required()
    def get(self, project_id, news_id):
        """
        Get news item
        ---
        description: Retrieves detailed information about a specific news item
          from a givenproject.
        tags:
          - News
        parameters:
          - in: path
            name: project_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the project
          - in: path
            name: news_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the news item
        responses:
          '200':
            description: News item successfully retrieved
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Unique news item identifier
                    title:
                      type: string
                      description: News item title
                    content:
                      type: string
                      description: News item content
                    created_at:
                      type: string
                      format: date-time
                      description: Creation timestamp
                    author_id:
                      type: string
                      format: uuid
                      description: Author's user ID
                    project_id:
                      type: string
                      format: uuid
                      description: Project identifier
          404:
            description: News item or project not found
        """
        projects_service.get_project(project_id)
        user_service.check_project_access(project_id)
        user_service.block_access_to_vendor()
        news = news_service.get_news(project_id, news_id)
        if len(news["data"]) > 0:
            return news["data"][0]
        else:
            raise NewsNotFoundException



#!/usr/bin/env python
import os
import sys
import tempfile

from zou.remote.config_payload import (
    check_config_version,
    get_config_from_payload,
    get_storage,
    get_file_from_storage,
    put_file_to_storage,
    make_key,
    setup_logging,
)

from zou.utils.movie import (
    normalize_movie,
    generate_thumbnail,
    generate_tile,
)
from zou.app.utils.thumbnail import (
    turn_into_thumbnail,
    generate_preview_variants,
)

logger = setup_logging()


def main():
    """
    Normalize given preview file in two videos, generate thumbnails
    and tiles.
    We assume that this script is launched in the context of a Nomad
    job. Arguments are retrieved from a payload file at JSON format.
    Nevertheless this script can be run directly from the command line.
    """
    config = get_config_from_payload()
    version = check_config_version(config)
    storage = get_storage(config, "movies")

    with tempfile.TemporaryDirectory() as tmpdir:
        preview_file_id = config["preview_file_id"]
        file_path = _fetch_movie_file(storage, tmpdir, preview_file_id)
        high_def_path, low_def_path, err = _run_normalize_movie(
            config, file_path
        )

        if err is None:
            put_file_to_storage(
                storage,
                high_def_path,
                make_key("previews", preview_file_id),
            )
            put_file_to_storage(
                storage,
                low_def_path,
                make_key("lowdef", preview_file_id),
            )
            logger.info(f"Normalization succeeded {high_def_path}")

            if version >= 2:
                pictures_storage = get_storage(config, "pictures")
                _generate_and_upload_thumbnails(
                    pictures_storage,
                    high_def_path,
                    preview_file_id,
                )
                _generate_and_upload_tile(
                    pictures_storage,
                    high_def_path,
                    preview_file_id,
                )
        else:
            logger.error(f"Normalization failed: {err}")
            sys.exit(1)


def _fetch_movie_file(storage, outdir, preview_file_id):
    """
    Download the movie file.
    """
    prefix = "source"
    filename = f"cache-previews-{preview_file_id}.mp4"
    file_path = os.path.join(outdir, filename)
    return get_file_from_storage(
        storage, file_path, make_key(prefix, preview_file_id)
    )


def _run_normalize_movie(config, movie_path):
    return normalize_movie(
        movie_path, config["fps"], config["width"], config["height"]
    )


def _generate_and_upload_thumbnails(
    pictures_storage, movie_path, preview_file_id
):
    """
    Generate thumbnail from movie, resize it, create variants
    (thumbnails, thumbnails-square, previews) and upload all to the
    pictures storage.
    """
    from zou.utils.movie import get_movie_size

    size = get_movie_size(movie_path)
    original_picture_path = generate_thumbnail(movie_path)
    turn_into_thumbnail(original_picture_path, size)

    variants = generate_preview_variants(
        original_picture_path, preview_file_id
    )
    variants.append(("original", original_picture_path))

    for prefix, path in variants:
        put_file_to_storage(
            pictures_storage, path, make_key(prefix, preview_file_id)
        )
        os.remove(path)
    logger.info(f"Thumbnails uploaded for {preview_file_id}")


def _generate_and_upload_tile(pictures_storage, movie_path, preview_file_id):
    """
    Generate tile mosaic from movie and upload to pictures storage.
    """
    try:
        tile_path = generate_tile(movie_path)
        put_file_to_storage(
            pictures_storage,
            tile_path,
            make_key("tiles", preview_file_id),
        )
        os.remove(tile_path)
        logger.info(f"Tile uploaded for {preview_file_id}")
    except Exception:
        logger.error("Failed to create tile", exc_info=True)


if __name__ == "__main__":
    main()



import semver
import shutil
from pathlib import Path

from zou.app import config, db
from zou.app.models.plugin import Plugin
from zou.app.utils.plugins import (
    PluginManifest,
    run_plugin_migrations,
    downgrade_plugin_migrations,
    uninstall_plugin_files,
    install_plugin_files,
    clone_git_repo,
)


def install_plugin(path, force=False):
    """
    Install a plugin: create folder, copy files, run migrations,
    and call pre/post install hooks.
    Supports local paths, zip files, and git repository URLs.
    """
    is_git_url = (
        path.startswith("http://")
        or path.startswith("https://")
        or path.startswith("git://")
        or path.startswith("ssh://")
        or path.startswith("git@")
    )

    temp_dir = None
    try:
        if is_git_url:
            cloned_path = clone_git_repo(path)
            temp_dir = cloned_path.parent
            path = cloned_path
        else:
            path = Path(path)
            if not path.exists():
                raise FileNotFoundError(
                    f"Plugin path '{path}' does not exist."
                )

        manifest = PluginManifest.from_plugin_path(path)
        plugin = Plugin.get_by(plugin_id=manifest.id)

        _run_plugin_hook(manifest.id, path, "pre_install", manifest)

        if plugin is not None:
            current = semver.Version.parse(plugin.version)
            new = semver.Version.parse(str(manifest.version))
            print(
                f"[Plugins] Upgrading {manifest.id}"
                f" from {current} to {new}..."
            )
            if not force and new <= current:
                print(
                    f"⚠️  [Plugins] Version {new} is not newer"
                    f" than {current}."
                )
            plugin.update(manifest.to_model_dict())
            print(f"[Plugins] Plugin {manifest.id} upgraded.")
        else:
            print(f"[Plugins] Installing plugin {manifest.id}...")
            plugin = Plugin.create(**manifest.to_model_dict())
            print(f"[Plugins] Plugin {manifest.id} installed.")

        print(
            f"[Plugins] Running database migrations"
            f" for {manifest.id}..."
        )
        plugin_path = install_plugin_files(
            path, Path(config.PLUGIN_FOLDER) / manifest.id
        )
        run_plugin_migrations(plugin_path, plugin)
        print(
            f"[Plugins] Database migrations for {manifest.id} applied."
        )

        # Re-query plugin instance after migrations
        # (Alembic operations may have detached it from the session)
        plugin = Plugin.get_by(plugin_id=manifest.id)

        _run_plugin_hook(
            manifest.id, plugin_path, "post_install", manifest
        )

        print_added_routes(plugin.plugin_id, plugin_path)
        return plugin.serialize()
    except Exception:
        print(
            f"❌ [Plugins] An error occurred while"
            f" installing/updating plugin..."
        )
        raise
    finally:
        if is_git_url and temp_dir and Path(temp_dir).exists():
            shutil.rmtree(temp_dir)


def uninstall_plugin(plugin_id):
    """
    Uninstall a plugin: call pre/post uninstall hooks, downgrade
    migrations, remove files, delete from database and remove folder.
    """
    print(f"[Plugins] Uninstalling plugin {plugin_id}...")
    plugin_path = Path(config.PLUGIN_FOLDER) / plugin_id

    manifest = None
    try:
        manifest = PluginManifest.from_plugin_path(plugin_path)
    except Exception:
        pass

    _run_plugin_hook(plugin_id, plugin_path, "pre_uninstall", manifest)

    downgrade_plugin_migrations(plugin_path)

    # Run post_uninstall before removing files (module must still exist)
    _run_plugin_hook(plugin_id, plugin_path, "post_uninstall", manifest)

    installed = uninstall_plugin_files(plugin_path)
    plugin = Plugin.query.filter_by(plugin_id=plugin_id).one_or_none()
    if plugin is not None:
        installed = True
        plugin.delete()

    if not installed:
        raise ValueError(f"Plugin '{plugin_id}' is not installed.")

    print(f"[Plugins] Plugin {plugin_id} uninstalled.")
    return True


def _import_plugin_module(plugin_id, plugin_path):
    """
    Import a plugin module dynamically from its install path.
    Returns the module or None if import fails.
    """
    import importlib
    import sys

    plugin_path = Path(plugin_path)
    plugin_folder = plugin_path.parent
    abs_plugin_path = str(plugin_folder.absolute())
    added_to_path = abs_plugin_path not in sys.path

    if added_to_path:
        sys.path.insert(0, abs_plugin_path)

    try:
        if plugin_id in sys.modules:
            return sys.modules[plugin_id]
        return importlib.import_module(plugin_id)
    except ImportError as e:
        print(f"⚠️  [Plugins] Could not import plugin module: {e}")
        return None
    except Exception:
        # Import can fail on reinstall (e.g. SQLAlchemy table already
        # defined). Return the partially-loaded module if available.
        if plugin_id in sys.modules:
            return sys.modules[plugin_id]
        return None
    finally:
        if added_to_path and abs_plugin_path in sys.path:
            sys.path.remove(abs_plugin_path)


def _run_plugin_hook(plugin_id, plugin_path, hook_name, *args):
    """
    Run a lifecycle hook (pre_install, post_install, pre_uninstall,
    post_uninstall) on a plugin module if it exists.
    Errors are logged but never block the install/uninstall flow.
    """
    try:
        plugin_module = _import_plugin_module(plugin_id, plugin_path)
        if plugin_module is None:
            return

        hook = getattr(plugin_module, hook_name, None)
        if hook is not None:
            print(f"[Plugins] Running {hook_name} for {plugin_id}...")
            hook(*args)
            print(f"[Plugins] {hook_name} for {plugin_id} completed.")
    except Exception as e:
        print(f"⚠️  [Plugins] {hook_name} failed for {plugin_id}: {e}")


def print_added_routes(plugin_id, plugin_path):
    """
    Print the added routes for a plugin.
    """
    print(f"[Plugins] Routes added by {plugin_id}:")

    plugin_module = _import_plugin_module(plugin_id, plugin_path)
    if plugin_module is not None and hasattr(plugin_module, "routes"):
        for route in plugin_module.routes:
            print(f"  - /plugins/{plugin_id}{route[0]}")
    else:
        print("  (No routes variable found in plugin)")

    print("--------------------------------")


def get_plugins():
    """
    Get all plugins.
    """
    return [plugin.present() for plugin in Plugin.query.all()]



from flask_jwt_extended import jwt_required

from zou.app.models.asset_instance import AssetInstance

from zou.app.services import assets_service, user_service
from zou.app.utils import permissions

from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource


class AssetInstancesResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, AssetInstance)

    @jwt_required()
    def get(self):
        """
        Get asset instances
        ---
        tags:
          - Crud
        description: Retrieve all asset instances. Supports filtering via
          query parameters and pagination.
        parameters:
          - in: query
            name: page
            required: false
            schema:
              type: integer
            example: 1
            description: Page number for pagination
          - in: query
            name: limit
            required: false
            schema:
              type: integer
            example: 50
            description: Number of results per page
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: false
            example: false
            description: Whether to include relations
        responses:
            200:
              description: Asset instances retrieved successfully
              content:
                application/json:
                  schema:
                    oneOf:
                      - type: array
                        items:
                          type: object
                      - type: object
                        properties:
                          data:
                            type: array
                            items:
                              type: object
                            example: []
                          total:
                            type: integer
                            example: 100
                          nb_pages:
                            type: integer
                            example: 2
                          limit:
                            type: integer
                            example: 50
                          offset:
                            type: integer
                            example: 0
                          page:
                            type: integer
                            example: 1
            400:
              description: Invalid filter format or query error
        """
        return super().get()

    @jwt_required()
    def post(self):
        """
        Create asset instance
        ---
        tags:
          - Crud
        description: Create a new asset instance with data provided in the
          request body. JSON format is expected.
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - asset_id
                properties:
                  asset_id:
                    type: string
                    format: uuid
                    example: a24a6ea4-ce75-4665-a070-57453082c25
                  name:
                    type: string
                    example: Instance Name
        responses:
            201:
              description: Asset instance created successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      asset_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        example: Instance Name
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid data format or validation error
        """
        return super().post()


class AssetInstanceResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, AssetInstance)
        self.protected_fields.append("number")

    @jwt_required()
    def get(self, instance_id):
        """
        Get asset instance
        ---
        tags:
          - Crud
        description: Retrieve an asset instance by its ID and return it as a
          JSON object. Supports including relations.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: true
            example: true
            description: Whether to include relations
        responses:
            200:
              description: Asset instance retrieved successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      asset_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        example: Instance Name
                      number:
                        type: integer
                        example: 1
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid ID format or query error
        """
        return super().get(instance_id)

    @jwt_required()
    def put(self, instance_id):
        """
        Update asset instance
        ---
        tags:
          - Crud
        description: Update an asset instance with data provided in the
          request body. JSON format is expected.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  name:
                    type: string
                    example: Updated Instance Name
        responses:
            200:
              description: Asset instance updated successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      asset_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      name:
                        type: string
                        example: Updated Instance Name
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T11:00:00Z"
            400:
              description: Invalid data format or validation error
        """
        return super().put(instance_id)

    @jwt_required()
    def delete(self, instance_id):
        """
        Delete asset instance
        ---
        tags:
          - Crud
        description: Delete an asset instance by its ID. Returns empty
          response on success.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
              description: Asset instance deleted successfully
            400:
              description: Integrity error or cannot delete
        """
        return super().delete(instance_id)

    def check_read_permissions(self, instance):
        if permissions.has_admin_permissions():
            return True
        else:
            asset_instance = self.get_model_or_404(instance["id"])
            asset = assets_service.get_asset(asset_instance.asset_id)
            user_service.check_project_access(asset["project_id"])
            user_service.check_entity_access(asset["id"])
            return True

    def check_update_permissions(self, asset_instance, data):
        if permissions.has_admin_permissions():
            return True
        else:
            asset = assets_service.get_asset(asset_instance["asset_id"])
            user_service.check_project_access(asset["project_id"])
            user_service.check_entity_access(asset["id"])
            return True



from flask_jwt_extended import jwt_required

from zou.app.models.entity import EntityLink
from zou.app.utils import fields

from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource
from zou.app.services.exception import (
    EntityLinkNotFoundException,
    WrongParameterException,
)


class EntityLinksResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, EntityLink)

    @jwt_required()
    def get(self):
        """
        Get entity links
        ---
        tags:
          - Crud
        description: Retrieve all entity links. Supports filtering via
          query parameters and pagination.
        parameters:
          - in: query
            name: page
            required: false
            schema:
              type: integer
            example: 1
            description: Page number for pagination
          - in: query
            name: limit
            required: false
            schema:
              type: integer
            example: 50
            description: Number of results per page
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: false
            example: false
            description: Whether to include relations
        responses:
            200:
              description: Entity links retrieved successfully
              content:
                application/json:
                  schema:
                    oneOf:
                      - type: array
                        items:
                          type: object
                      - type: object
                        properties:
                          data:
                            type: array
                            items:
                              type: object
                            example: []
                          total:
                            type: integer
                            example: 100
                          nb_pages:
                            type: integer
                            example: 2
                          limit:
                            type: integer
                            example: 50
                          offset:
                            type: integer
                            example: 0
                          page:
                            type: integer
                            example: 1
            400:
              description: Invalid filter format or query error
        """
        return super().get()

    @jwt_required()
    def post(self):
        """
        Create entity link
        ---
        tags:
          - Crud
        description: Create a new entity link with data provided in the
          request body. JSON format is expected. Links entities together
          in casting relationships.
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - entity_in_id
                  - entity_out_id
                properties:
                  entity_in_id:
                    type: string
                    format: uuid
                    example: a24a6ea4-ce75-4665-a070-57453082c25
                  entity_out_id:
                    type: string
                    format: uuid
                    example: b24a6ea4-ce75-4665-a070-57453082c25
                  nb_occurences:
                    type: integer
                    default: 1
                    example: 1
                  label:
                    type: string
                    example: fixed
        responses:
            201:
              description: Entity link created successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      entity_in_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      entity_out_id:
                        type: string
                        format: uuid
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      nb_occurences:
                        type: integer
                        example: 1
                      label:
                        type: string
                        example: fixed
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid data format or validation error
        """
        return super().post()


class EntityLinkResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, EntityLink)

    @jwt_required()
    def get(self, instance_id):
        """
        Get entity link
        ---
        tags:
          - Crud
        description: Retrieve an entity link by its ID and return it as a
          JSON object. Supports including relations.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: true
            example: true
            description: Whether to include relations
        responses:
            200:
              description: Entity link retrieved successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      entity_in_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      entity_out_id:
                        type: string
                        format: uuid
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      nb_occurences:
                        type: integer
                        example: 1
                      label:
                        type: string
                        example: fixed
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid ID format or query error
        """
        return super().get(instance_id)

    @jwt_required()
    def put(self, instance_id):
        """
        Update entity link
        ---
        tags:
          - Crud
        description: Update an entity link with data provided in the
          request body. JSON format is expected.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  nb_occurences:
                    type: integer
                    example: 2
                  label:
                    type: string
                    example: updated
        responses:
            200:
              description: Entity link updated successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      entity_in_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      entity_out_id:
                        type: string
                        format: uuid
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      nb_occurences:
                        type: integer
                        example: 2
                      label:
                        type: string
                        example: updated
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T11:00:00Z"
            400:
              description: Invalid data format or validation error
        """
        return super().put(instance_id)

    @jwt_required()
    def delete(self, instance_id):
        """
        Delete entity link
        ---
        tags:
          - Crud
        description: Delete an entity link by its ID. Returns empty
          response on success.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
              description: Entity link deleted successfully
            400:
              description: Integrity error or cannot delete
        """
        return super().delete(instance_id)

    def get_model_or_404(self, instance_id):
        if not fields.is_valid_id(instance_id):
            raise WrongParameterException("Malformed ID.")
        instance = self.model.get_by(id=instance_id)
        if instance is None:
            raise EntityLinkNotFoundException
        return instance
